"""Classes used by the Inputs class"""
import os
import numpy as np
import pickle
from astropy.time import Time
import astropy.units as u
from sqlalchemy import text
from nexoclom.solarsystem import SSObject
from nexoclom.utilities import InputError, NexoclomConfig


dtor = np.pi/180.
# Tolerances for floating point values
dtaa = 2.*dtor


class InputClass:
    def __init__(self):
        self.type = None
        self.config = NexoclomConfig()
    
    def __repr__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'geometry.{key} = {value}\n'
        return result.strip()
    
    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])
        
    def make_query(self):
        fields = ', '.join(self.__dict__.keys())
        values = '%' + ', %'.join(self.__dict__.keys())
        return fields, values

    def insert(self):
        fields, values = self.make_query()
        query = text(f'''INSERT INTO {self.type} ({fields})
                         VALUES ({values});''')
        with self.config.create_engine().begin as con:
            con.execute(query, self.__dict__)
            
        ids = self.search()
        return ids
    
    def search(self):
        fields, values = self.make_query()
        query = text(f'''SELECT idnum FROM {self.type}
                         WHERE {fields}''')
        with self.config.create_engine().connect() as con:
            result = con.execute(query, self.__dict__)
            ids = [row.idnum for row in result]
        return ids
        

class Geometry(InputClass):
    def __init__(self, gparam):
        """Define a Geometry object.
        
        See :doc:`inputfiles#Geometry` for more information.
        """
        # Planet
        super().__init__()
        
        planet = gparam.get('planet', None)
        if planet is None:
            raise InputError('Geometry.__init__',
                             'Planet not defined in inputfile.')
        else:
            self.planet = SSObject(planet.title())

        # All possible objects
        objlist = [self.planet.object]
        if self.planet.moons is not None:
            objlist.extend([m.object for m in self.planet.moons])
        else:
            pass

        # Choose the starting point
        self.startpoint = gparam.get('startpoint', self.planet.object).title()
        if self.startpoint not in objlist:
            print(f'{self.startpoint} is not a valid starting point.')
            olist = '\n\t'.join(objlist)
            print(f'Valid choices are:\n\t{olist}')
            raise ValueError
        else:
            pass

        # Choose which objects to include
        # This is given as a list of names
        # Default = geometry.planet and geometry.startpoint
        if 'objects' in gparam:
            inc = set(i.strip().title()
                      for i in gparam['objects'].split(','))
        else:
            inc = {self.planet.object, self.startpoint}

        for i in inc:
            if i not in objlist:
                raise InputError('Geometry.__init__',
                                 f'Invalid object {i} in geometry.include')
            
        # Only remember objects that will be included
        self.objects = set(SSObject(o) for o in inc)
        if len(self.objects) == 0:
            # Probably not possible to get here
            self.objects = None
        else:
            pass

        # Different objects are created for geometry_with_starttime and
        # geometry_without_starttime
        if 'starttime' in gparam:
            self.type = 'geometry with starttime'
            try:
                self.time = Time(gparam['starttime'].upper())
            except:
                raise InputError('Geometry.__init__',
                         f'Invalid starttime format: {gparam["starttime"]}')
        else:
            self.type = 'geometry without starttime'
            if len(self.planet) == 1:
                self.phi = None
            elif 'phi' in gparam:
                phi_ = gparam['phi'].split(',')
                phi = tuple(float(p)*u.rad for p in phi_)
                nmoons = len(self.objects - {self.planet})
                if len(phi) == nmoons:
                    self.phi = phi
                else:
                    raise InputError('Geometry.__init__',
                          'The wrong number of orbital positions was given.')
            else:
                raise InputError('Geometry.__init__',
                                 'geometry.phi was not specified.')

            # Subsolar longitude and latitude
            if 'subsolarpoint' in gparam:
                subs = gparam['subsolarpoint'].split(',')
                try:
                    self.subsolarpoint = (float(subs[0])*u.rad,
                                          float(subs[1])*u.rad)
                except:
                    raise InputError('Geometry.__init__',
                         'The format for geometry.subsolarpoint is wrong.')
            else:
                self.subsolarpoint = (0*u.rad, 0*u.rad)

            # True Anomaly Angle
            self.taa = float(gparam.get('taa', 0.))*u.rad
            

class SurfaceInteraction:
    def __init__(self, sparam):
        """Define a SurfaceInteraction object.

        See :doc:`inputfiles#SurfaceInteraction` for more information.
        """
        sticktype = (sparam['sticktype'].lower()
                     if 'sticktype' in sparam
                     else None)
        if sticktype == 'temperature dependent':
            self.sticktype = sticktype
            
            if 'accomfactor' in sparam:
                self.accomfactor = float(sparam['accomfactor'])
            else:
                raise InputError('SurfaceInteraction.__init__',
                                 'surfaceinteraction.accomfactor not given.')
            
            if 'a' in sparam:
                A = tuple([float(a) for a in sparam['a'].split(',')])
                if len(A) == 3:
                    self.A = A
                else:
                    raise InputError('SurfaceInteraction.__init__',
                                     'surfaceinteraction.A must have 3 values')
            else:
                self.A = (1.57014, -0.006262, 0.1614157)
        elif sticktype == 'surface map':
            self.sticktype = sticktype
            self.stick_mapfile = sparam.get('stick_mapfile', 'default')

            if os.path.exists(self.stick_mapfile):
                if self.stick_mapfile.endswith('.pkl'):
                    with open(self.stick_mapfile, 'rb') as f:
                        sourcemap = pickle.load(f)
                    self.coordinate_system = sourcemap.get('coordinate_system',
                                                           'solar-fixed')
                else:
                    self.coordinate_system = 'solar-fixed'
                    # assert 0, 'Not set up yet'
            else:
                print('Warning: mapfile does not exist')
                self.coordinate_system = 'solar-fixed'

            if (self.coordinate_system == 'planet-fixed'):
                self.subsolarlon = sparam.get('subsolarlon', 0.) * u.rad
            else:
                self.subsolarlon = None

            if 'accomfactor' in sparam:
                self.accomfactor = float(sparam['accomfactor'])
            else:
                raise InputError('SurfaceInteraction.__init__',
                                 'surfaceinteraction.accomfactor not given.')
        elif 'stickcoef' in sparam:
            # Constant sticking
            self.sticktype = 'constant'
            self.stickcoef = float(sparam['stickcoef'])
            if self.stickcoef < 0:
                self.stickcoef = 0
            elif self.stickcoef > 1:
                self.stickcoef = 1
            else:
                pass
            
            if 'accomfactor' in sparam:
                self.accomfactor = float(sparam['accomfactor'])
            else:
                if self.stickcoef == 1:
                    self.accomfactor = None
                else:
                    raise InputError('SurfaceInteraction.__init__',
                                 'surfaceinteraction.accomfactor not given.')
        else:
            self.sticktype = 'constant'
            self.stickcoef = 1.
            self.accomfactor = None

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'surfaceinteraction.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            if self.sticktype == 'constant':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO surface_int_constant (
                                       stickcoef, accomfactor) VALUES (
                                       %s, %s)''',
                                (self.stickcoef, self.accomfactor))
            elif self.sticktype == 'surface map':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO surface_int_map (
                                       mapfile, accomfactor) VALUES (
                                       %s, %s)''',
                                (self.stick_mapfile, self.accomfactor))
            elif self.sticktype == 'temperature dependent':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO surface_int_tempdependent (
                                       accomfactor, a) VALUES
                                       (%s, %s::DOUBLE PRECISION[3])''',
                                (self.accomfactor, self.A))
            else:
                raise InputError('SurfaceInteraction.search()',
                    f'surfaceinteraction.sticktype = {self.sticktype} not allowed.')
            ids = self.search()
            assert ids is not None
        else:
            pass

        return ids

    def search(self):
        if self.sticktype == 'constant':
            params = [self.stickcoef]
            if self.accomfactor is None:
                afactor = 'accomfactor is NULL'
            else:
                afactor = 'accomfactor = %s'
                params.append(self.accomfactor)
            
            query = f"""SELECT idnum
                        FROM surface_int_constant
                        WHERE stickcoef = %s and
                              {afactor}"""
        elif self.sticktype == 'temperature dependent':
            params = [self.accomfactor, self.A]
            query = f"""SELECT idnum
                        FROM surface_int_tempdependent
                        WHERE accomfactor = %s and
                              a = %s::DOUBLE PRECISION[3]"""
        elif self.sticktype == 'surface map':
            params = [self.stick_mapfile, self.accomfactor]
            query = f"""SELECT idnum
                        FROM surface_int_map
                        WHERE mapfile = %s and
                              accomfactor = %s"""
        else:
            raise InputError('SurfaceInteraction.search()',
             f'surfaceinteraction.sticktype = {self.sticktype} not allowed.')
        
        with config.database_connect() as con:
            cur = con.cursor()
            cur.execute(query, tuple(params))
    
            if cur.rowcount == 0:
                return None
            elif cur.rowcount == 1:
                return cur.fetchone()[0]
            else:
                raise RuntimeError('SurfaceInteraction.search()',
                                   'Duplicates in surface interaction table')


class Forces:
    def __init__(self, fparam):
        """Define a Forces object.

        See :doc:`inputfiles#Forces` for more information.
        """
    
        self.gravity = (bool(eval(fparam['gravity'].title()))
                        if 'gravity' in fparam
                        else True)
        self.radpres = (bool(eval(fparam['radpres'].title()))
                        if 'radpres' in fparam
                        else True)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'forces.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            with config.database_connect() as con:
                cur = con.cursor()
                cur.execute('''INSERT INTO forces (
                                   gravity, radpres) VALUES (%s, %s)''',
                            (self.gravity, self.radpres))
            ids = self.search()
            assert ids is not None
        else:
                pass

        return ids

    def search(self):
        query = f'''SELECT idnum FROM forces
                    WHERE gravity = %s and
                          radpres = %s'''
        
        with config.database_connect() as con:
            cur = con.cursor()
            cur.execute(query, (self.gravity, self.radpres))
    
            if cur.rowcount == 0:
                return None
            elif cur.rowcount == 1:
                return cur.fetchone()[0]
            else:
                raise RuntimeError('Forces.search()',
                                   'Duplicates in forces table')


class SpatialDist:
    def __init__(self, sparam):
        """Define a SpatialDist object.

        See :doc:`inputfiles#SpatialDist` for more information.
        """
        if 'type' in sparam:
            self.type = sparam['type']
        else:
            raise InputError('SpatialDist.__init__',
                             'SpatialDist.type not given')
        
        if self.type == 'uniform':
            self.exobase = (float(sparam['exobase'])
                            if 'exobase' in sparam
                            else 1.)  # Unit gets set later
            if 'longitude' in sparam:
                lon0, lon1 = (float(l.strip())
                              for l in sparam['longitude'].split(','))
                lon0 = max(lon0, 0.)
                lon0 = min(lon0, 2*np.pi)
                lon1 = max(lon1, 0.)
                lon1 = min(lon1, 2*np.pi)
                self.longitude = (lon0*u.rad, lon1*u.rad)
            else:
                self.longitude = (0.*u.rad, 2*np.pi*u.rad)
                
            if 'latitude' in sparam:
                lat0, lat1 = (float(l.strip())
                              for l in sparam['latitude'].split(','))
                lat0 = max(lat0, -np.pi/2)
                lat0 = min(lat0, np.pi/2)
                lat1 = max(lat1, -np.pi/2)
                lat1 = min(lat1, np.pi/2)
                if lat0 > lat1:
                    raise InputError('SpatialDist.__init__',
                         'SpatialDist.latitude[0] > SpatialDist.latitude[1]')
                self.latitude = (lat0*u.rad, lat1*u.rad)
            else:
                self.latitude = (-np.pi/2*u.rad, np.pi/2*u.rad)
        elif self.type == 'surface map':
            self.exobase = (float(sparam['exobase'])
                            if 'exobase' in sparam
                            else 1.)  # Unit gets set later
            
            self.mapfile = sparam.get('mapfile', 'default')
            
            if os.path.exists(self.mapfile):
                if self.mapfile.endswith('.pkl'):
                    with open(self.mapfile, 'rb') as f:
                        sourcemap = pickle.load(f)
                    coord_systems = ['planet-fixed', 'solar-fixed']
                    assert sourcemap['coordinate_system'] in coord_systems, (
                        'source map must be in planet-fixed or solar-fixed coordinates.')
                    self.coordinate_system = sourcemap['coordinate_system']
                else:
                    self.coordinate_system = 'solar-fixed'
                    # assert 0, 'Not set up yet'
            else:
                print('Warning: mapfile does not exist')
                self.coordinate_system = 'planet-fixed'

            if (self.coordinate_system == 'planet-fixed'):
                self.subsolarlon = sparam.get('subsolarlon', 0.)*u.rad
            else:
                self.subsolarlon = None
        elif self.type == 'surface spot':
            self.exobase = (float(sparam['exobase'])
                            if 'exobase' in sparam
                            else 1.)  # Unit gets set later
            if 'longitude' in sparam:
                self.longitude = float(sparam['longitude'])*u.rad
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpatialDist.longitude not given.')
            
            if 'latitude' in sparam:
                self.latitude = float(sparam['latitude'])*u.rad
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpatialDist.latitude not given.')

            if 'sigma' in sparam:
                self.sigma = float(sparam['sigma'])*u.rad
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpatialDist.sigma not given.')
        elif self.type == 'fitted output':
            self.unfit_outid = -1
            self.query = None
        else:
            raise InputError('SpatialDist.__init__',
                             f'SpatialDist.type = {self.type} not defined.')

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'SpatialDist.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            if self.type == 'uniform':
                long = [l.value for l in self.longitude]
                lat = [l.value for l in self.latitude]
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO spatdist_uniform (
                                       exobase, longitude, latitude) VALUES (
                                       %s, %s::DOUBLE PRECISION[2],
                                       %s::DOUBLE PRECISION[2])''',
                                (self.exobase, long, lat))
            elif self.type == 'surface map':
                with config.database_connect() as con:
                    cur = con.cursor()
                    sslon = (None
                             if self.subsolarlon is None
                             else self.subsolarlon.value)
                    cur.execute('''INSERT INTO spatdist_surfmap (
                                       exobase, mapfile, subsolarlon,
                                       coordinate_system) VALUES (
                                       %s, %s, %s, %s)''',
                                (self.exobase, self.mapfile, sslon,
                                 self.coordinate_system))
            elif self.type == 'surface spot':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO spatdist_spot (
                                       exobase, longitude, latitude, sigma) VALUES (
                                       %s, %s, %s, %s)''',
                                (self.exobase, self.longitude.value,
                                 self.latitude.value, self.sigma.value))
            elif self.type == 'fitted output':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO spatdist_fittedoutput (
                                   unfit_outid, query)
                                       VALUES (%s, %s)''',
                                (self.unfit_outid, self.query))
            else:
                raise InputError('SpatialDist.search()',
                                 f'SpatialDist.type = {self.type} not allowed.')
            ids = self.search()
            assert ids is not None
        else:
            pass
        
        return ids

    def search(self):
        if self.type == 'uniform':
            params = [self.exobase,
                      [self.longitude[0].value, self.longitude[1].value],
                      [self.latitude[0].value, self.latitude[1].value]]
            query = '''SELECT idnum
                       FROM spatdist_uniform
                       WHERE exobase = %s and
                             longitude = %s::DOUBLE PRECISION[2] and
                             latitude = %s::DOUBLE PRECISION[2]'''
        elif self.type == 'surface map':
            params = [self.exobase, self.mapfile]
            if self.subsolarlon is None:
                sslon = 'subsolarlon is NULL'
            else:
                sslon = 'subsolarlon = %s'
                params.append(self.subsolarlon.value)
                
            params.append(self.coordinate_system)
            query = f'''SELECT idnum
                        FROM spatdist_surfmap
                        WHERE exobase = %s and
                              mapfile = %s and
                              {sslon} and
                              coordinate_system = %s'''
        elif self.type == 'surface spot':
            params = [self.exobase, self.longitude.value, self.latitude.value,
                      self.sigma.value]
            query = '''SELECT idnum
                       FROM spatdist_spot
                       WHERE exobase = %s and
                             longitude = %s and
                             latitude = %s and
                             sigma = %s'''
        elif self.type == 'fitted output':
            params = [self.unfit_outid, self.query]
            query = '''SELECT idnum
                       FROM spatdist_fittedoutput
                       WHERE unfit_outid = %s and
                             query = %s'''
        else:
            raise InputError('SpatialDist.__init__',
                             f'SpatialDist.type = {self.type} not defined.')

        with config.database_connect() as con:
            cur = con.cursor()
            # print(cur.mogrify(query, tuple(params)))
            cur.execute(query, tuple(params))
    
            if cur.rowcount == 0:
                return None
            elif cur.rowcount == 1:
                return cur.fetchone()[0]
            else:
                raise RuntimeError('SpatialDist.search()',
                                   'Duplicates in spatial distribution table')
        


class SpeedDist:
    """Define a SpeedDist object.

    See :doc:`inputfiles#SpeedDist` for more information.
    """
    def __init__(self, sparam):
        self.type = sparam['type']

        if self.type == 'gaussian':
            if 'vprob' in sparam:
                self.vprob = float(sparam['vprob'])*u.km/u.s
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.vprob not given.')
            if 'sigma' in sparam:
                self.sigma = float(sparam['sigma'])*u.km/u.s
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.sigma not given.')
        elif self.type == 'sputtering':
            if 'alpha' in sparam:
                self.alpha = float(sparam['alpha'])
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.alpha not given.')
            if 'beta' in sparam:
                self.beta = float(sparam['beta'])
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.beta not given.')
            if 'u' in sparam:
                self.U = float(sparam['u'])*u.eV
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.U not given.')
        elif self.type == 'maxwellian':
            if 'temperature' in sparam:
                self.temperature = float(sparam['temperature'])*u.K
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.temperature not given.')
        elif self.type == 'flat':
            if 'vprob' in sparam:
                self.vprob = float(sparam['vprob'])*u.km/u.s
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.vprob not given.')
            
            if 'delv' in sparam:
                self.delv = float(sparam['delv'])*u.km/u.s
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.delv not given.')
        elif self.type == 'user defined':
            self.vdistfile = sparam.get('vdistfile', 'default')
        elif self.type == 'fitted output':
            self.unfit_outid = -1
            self.query = None
        else:
            assert 0, f'SpeedDist.type = {self.type} not available'

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'SpeedDist.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            if self.type == 'gaussian':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO speeddist_gaussian (
                                       vprob, sigma) VALUES (%s, %s)''',
                                (self.vprob.value, self.sigma.value))
            elif self.type == 'sputtering':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO speeddist_sputtering (
                                       alpha, beta, U) VALUES (%s, %s, %s)''',
                                (self.alpha, self.beta, self.U.value))
            elif self.type == 'maxwellian':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO speeddist_maxwellian (
                                       temperature) VALUES (%s)''',
                                (self.temperature.value, ))
            elif self.type == 'flat':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO speeddist_flat (
                                       vprob, delv) VALUES (%s, %s)''',
                                (self.vprob.value, self.delv.value))
            elif self.type == 'user defined':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO speeddist_user (
                                       vdistfile) VALUES (%s)''',
                                (self.vdistfile,))
            elif self.type == 'fitted output':
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO speeddist_fittedoutput (
                                   unfit_outid, query)
                                   VALUES (%s, %s)''',
                                (self.unfit_outid, self.query))
            else:
                raise InputError('SpeedDist.search()',
                                 f'speeddist.type = {self.type} not allowed.')

            ids = self.search()
            assert ids is not None
        else:
            pass

        return ids

    def search(self):
        if self.type == 'gaussian':
            params = [self.vprob.value, self.sigma.value]
            query = '''SELECT idnum
                       FROM speeddist_gaussian
                       WHERE vprob = %s and
                             sigma = %s'''
        elif self.type == 'sputtering':
            params = [self.alpha, self.beta, self.U.value]
            query = '''SELECT idnum
                       FROM speeddist_sputtering
                       WHERE alpha = %s and
                             beta = %s and
                             U = %s'''
        elif self.type == 'maxwellian':
            params = [self.temperature.value]
            query = '''SELECT idnum
                       FROM speeddist_maxwellian
                       WHERE temperature = %s'''
        elif self.type == 'flat':
            params = [self.vprob.value, self.delv.value]
            query = '''SELECT idnum
                       FROM speeddist_flat
                       WHERE vprob = %s and
                             delv = %s'''
        elif self.type == 'user defined':
            params = [self.vdistfile]
            query = '''SELECT idnum
                       FROM speeddist_user
                       WHERE vdistfile = %s'''
        elif self.type == 'fitted output':
            params = [self.unfit_outid, self.query]
            query = '''SELECT idnum
                           FROM speeddist_fittedoutput
                           WHERE unfit_outid = %s and
                                 query = %s'''
        else:
            raise InputError('SpeedDist.__init__',
                             f'SpeedDist.type = {self.type} not defined.')

        with config.database_connect() as con:
            cur = con.cursor()
            cur.execute(query, tuple(params))
        
            if cur.rowcount == 0:
                return None
            elif cur.rowcount == 1:
                return cur.fetchone()[0]
            else:
                raise RuntimeError('SpeedDist.search()',
                                   'Duplicates in speed distribution table')
    

class AngularDist:
    def __init__(self, aparam):
        """Define a AngularDist object.

        See :doc:`inputfiles#AngularDist` for more information.
        """
        if 'type' in aparam:
            self.type = aparam['type']
            if self.type == 'radial':
                pass
            elif self.type == 'isotropic':
                if 'azimuth' in aparam:
                    az0, az1 = (float(l.strip())
                                for l in aparam['azimuth'].split(','))
                    az0 = max(az0, 0.)
                    az0 = min(az0, 2*np.pi)
                    az1 = max(az1, 0.)
                    az1 = min(az1, 2*np.pi)
                    self.aziumth = (az0*u.rad, az1*u.rad)
                else:
                    self.azimuth = (0*u.rad, 2*np.pi*u.rad)

                if 'altitude' in aparam:
                    alt0, alt1 = (float(l.strip())*u.rad
                                  for l in aparam['altitude'].split(','))
                    alt0 = max(alt0, 0)
                    alt0 = min(alt0, np.pi/2)
                    alt1 = max(alt1, 0)
                    alt1 = min(alt1, np.pi/2)
                    if alt0 > alt1:
                        raise InputError('AngularDist.__init__',
                         'AngularDist.altitude[0] > AngularDist.altitude[1]')
                    self.altitude = (alt0*u.rad, alt1*u.rad)
                else:
                    self.altitude = (0*u.rad, np.pi/2*u.rad)
            else:
                raise InputError('AngularDist.__init__',
                             f'AngularDist.type = {self.type} not defined.')
        else:
            self.type = 'isotropic'
            self.azimuth = (0*u.rad, 2*np.pi*u.rad)
            self.altitude = (0*u.rad, np.pi/2*u.rad)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'AngularDist.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            if self.type == 'radial':
                assert 0, 'Should not be able to get here.'
            elif self.type == 'isotropic':
                alt = [a.value for a in self.altitude]
                az = [a.value for a in self.azimuth]
                with config.database_connect() as con:
                    cur = con.cursor()
                    cur.execute('''INSERT INTO angdist_isotropic (
                                       altitude, azimuth) VALUES (
                                       %s::DOUBLE PRECISION[2],
                                       %s::DOUBLE PRECISION[2])''',
                                (alt, az))
            else:
                raise InputError('AngularDist.search()',
                                 f'angulardist.type = {self.type} not allowed.')

            ids = self.search()
            assert ids is not None
        else:
            pass

        return ids

    def search(self):
        if self.type == 'radial':
            return 0
        elif self.type == 'isotropic':
            params = [[a.value for a in self.altitude],
                      [a.value for a in self.azimuth]]
            query = '''SELECT idnum
                       FROM angdist_isotropic
                       WHERE altitude=%s::DOUBLE PRECISION[2] and
                             azimuth=%s::DOUBLE PRECISION[2]'''
        else:
            raise InputError('AngularDist.__init__',
                             f'AngularDist.type = {self.type} not defined.')
    
        with config.database_connect() as con:
            cur = con.cursor()
            cur.execute(query, tuple(params))
        
            if cur.rowcount == 0:
                return None
            elif cur.rowcount == 1:
                return cur.fetchone()[0]
            else:
                raise RuntimeError('AngularDist.search()',
                                   'Duplicates in angular distribution table')


class Options:
    def __init__(self, oparam):
        """Define a Options object.

        See :doc:`inputfiles#Options` for more information.
        """
        if 'endtime' in oparam:
            self.endtime = float(oparam['endtime'])*u.s
        else:
            raise InputError('Options.__init__',
                             'options.endtime not specified.')

        if 'species' in oparam:
            self.species = oparam['species'].capitalize()
        elif 'atom' in oparam:
            self.species = oparam['atom'].capitalize()
        else:
            raise InputError('Options.__init__',
                             'options.species not specified.')

        self.lifetime = float(oparam.get('lifetime', 0))*u.s
        
        if 'outeredge' in oparam:
            self.outeredge = float(oparam['outeredge'])
        elif 'outer_edge' in oparam:
            self.outeredge = float(oparam['outer_edge'])
        else:
            self.outeredge = 1e30
            
        if 'step_size' in oparam:
            self.step_size = float(oparam['step_size'])
        elif 'stepsize' in oparam:
            self.step_size = float(oparam['step_size'])
        else:
            self.step_size = 0.

        if self.step_size == 0:
            self.resolution = oparam.get('resolution', 1e-4)
        else:
            self.resolution = None
            
        if 'fitted' in oparam:
            self.fitted = (True if oparam['fitted'].casefold() == 'True'.casefold()
                           else False)
        else:
            self.fitted = False

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'options.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            with config.database_connect() as con:
                cur = con.cursor()
                cur.execute('''INSERT into options (endtime, species, lifetime,
                                   outer_edge, step_size, resolution, fitted) VALUES (
                                   %s, %s, %s, %s, %s, %s, %s)''',
                            (self.endtime.value, self.species,
                             self.lifetime.value, self.outeredge,
                             self.step_size, self.resolution, self.fitted))
            ids = self.search()
            assert ids is not None
        else:
            pass

        return ids

    def search(self):
        params = [self.endtime.value, self.species, self.lifetime.value,
                  self.outeredge, self.step_size, self.fitted]
        
        if self.resolution is None:
            resol = 'resolution is NULL'
        else:
            resol = 'resolution = %s'
            params.append(self.resolution)
            
        query = f'''SELECT idnum
                    FROM options
                    WHERE endtime = %s and
                          species = %s and
                          lifetime = %s and
                          outer_edge = %s and
                          step_size = %s and
                          {resol} and
                          fitted = %s'''
        
        with config.database_connect() as con:
            cur = con.cursor()
            cur.execute(query, tuple(params))
    
            if cur.rowcount == 0:
                return None
            elif cur.rowcount == 1:
                return cur.fetchone()[0]
            else:
                raise RuntimeError('Options.search()',
                                   'Duplicates in options table')



import os.path
import numpy as np
import pandas as pd
import pickle
import random
import copy
import astropy.units as u
from astropy.modeling import models, fitting
from astropy.visualization import PercentileInterval
from sklearn.neighbors import KDTree, BallTree
import sqlalchemy as sqla
import sqlalchemy.dialects.postgresql as pg
from nexoclom.modelcode.ModelResult import ModelResult
from nexoclom.modelcode.Output import Output
from nexoclom.modelcode.input_classes import SpatialDist, SpeedDist

xcols = ['x', 'y', 'z']
borecols = ['xbore', 'ybore', 'zbore']


class IterationResult:
    def __init__(self, iteration):
        self.radiance = iteration['radiance']
        self.npackets = iteration['npackets']
        self.totalsource = iteration['totalsource']
        self.outputfile = iteration['outputfile']
        self.out_idnum = iteration['out_idnum']
        self.modelfile = None
        self.model_idnum = None
        self.fitted = False
        
class FittedIterationResult(IterationResult):
    def __init__(self, iteration):
        super().__init__(iteration)
        
        self.unfit_outputfile = iteration['unfit_outputfile']
        self.unfit_outid = iteration['unfit_outid']
        self.fitted = True
        self.saved_packets = None
        self.weighting = None
        self.included = None

class LOSResult(ModelResult):
    """Class to contain the LOS result from multiple outputfiles.
    
    Determine column or emission along lines of sight.
    This assumes the model has already been run.

    **Parameters**
    
    scdata
        Spacecraft data object (currently designed for MESSENGERdata object
        but can be faked for other types of data)

    params
        A dictionary containing the keys
        
            * quantity [required]: column, density, radiance
            
            * wavelength [optional]: For radiance, wavelenghts to be simulated.
            If not given, uses defaults for species. Must be a valid emission
            line for the species.
            
        More parameters will be added when more emission processes are included.
        For now, the easiest is `params = {'format': 'radiance'}`

    dphi
        Angular size of the view cone. Default = r deg.
        
    **Methods**
    
    **Attributes**
   
    species, query
        The species and query used to retrieve the data used. These can be
        used to retrieve the data if necessary
        
    type
        'LineOfSight' for a line of sight result
        
    dphi
        boresight opening angle
        
    radiance
        Pandas series containing modeled radiance along each line of sight
        
    npackets
        Pandas series containing the number of packets along each line of sight
    
    sourcemap
        Characterization of the initial source (spatial and velocity distributions)
        
    modelfiles
        Saved LOS Iteration results
    
    _oedge
        Maximum distance from the s/c to integrate. Twice the outer edge of the
        simulation region or 100 R_planet, whichever is less.
    """
    def __init__(self, scdata, inputs, params=None, dphi=1*u.deg, **kwargs):
        """Initializes the LOSResult and runs the model if necessary"""
        if params is None:
            params = {'quantity': 'radiance'}
        else:
            pass

        scdata.set_frame('Model')
        super().__init__(inputs, params)
        
        # Basic information
        self.species = scdata.species
        self.query = scdata.query
        self.type = 'LineOfSight'
        self.dphi = dphi.to(u.rad).value
        self._oedge = np.min([self.inputs.options.outeredge*2, 100])

        self.fitted = self.inputs.options.fitted
        nspec = len(scdata)
        self.radiance = pd.Series(np.zeros(nspec), index=scdata.data.index)
        self.radiance_unit = u.def_unit('kR', 1e3*u.R)
        self.sourcemap = None
        self.modelfiles = None
        
        self.goodness_of_fit = None
        self.mask = None
        self.masking = kwargs.get('masking', None)
        self.fit_method = kwargs.get('fit_method', None)
        self.label = kwargs.get('label', 'LOSResult')

        if self.fitted:
            self.unfit_outid = None
            self.unfit_outputfiles = None
        else:
            pass
        
    def __repr__(self):
        return self.__str__()
        
    def __str__(self):
        return f'''quantity = {self.quantity}
npackets = {self.npackets}
totalsource = {self.totalsource}
atoms per packet = {self.atoms_per_packet}
sourcerate = {self.sourcerate}
dphi = {self.dphi}
fit_method = {self.fit_method}
fitted = {self.fitted}'''
    
    # Helper functions
    @staticmethod
    def _should_add_weight(index, saved):
        return index in saved

    @staticmethod
    def _add_weight(x, ratio):
        return np.append(x, ratio)

    @staticmethod
    def _add_index(x, i):
        return np.append(x, i)

    def save(self, iteration_result):
        '''
        Insert the result of a LOS iteration into the database
        :param iteration_result: LOS result from a single outputfile
        :return: name of saved file
        '''
        if isinstance(iteration_result, FittedIterationResult):
            ufit_id = iteration_result.unfit_outid
        else:
            ufit_id = None

        # Insert the result into the database
        engine = self.inputs.config.create_engine()
        metadata_obj = sqla.MetaData()
        table = sqla.Table("uvvsmodels", metadata_obj, autoload_with=engine)
        
        insert_stmt = pg.insert(table).values(
            out_idnum = iteration_result.out_idnum,
            unfit_idnum = ufit_id,
            quantity = self.quantity,
            query = self.query,
            dphi = self.dphi,
            mechanism = self.mechanism,
            wavelength = [w.value for w in self.wavelength],
            fitted = self.fitted)
        
        with engine.connect() as con:
            result = con.execute(insert_stmt)
            con.commit()
            
        self.idnum = result.inserted_primary_key[0]
        savefile = os.path.join(os.path.dirname(iteration_result.outputfile),
                                f'model.{self.idnum}.pkl')
        print(f'Saving model result {savefile}')
        update = sqla.update(table).where(table.columns.idnum == self.idnum).values(
            filename=savefile)
        with engine.connect() as con:
            con.execute(update)
            con.commit()

        with open(savefile, 'wb') as f:
            pickle.dump(iteration_result, f)
        
        return savefile
    
    def search(self):
        """
        :return: dictionary containing search results:
                 {outputfilename: (modelfile_id, modelfile_name)}
        """
        search_results = {}
        for oid, outputfile in zip(self.outid, self.outputfiles):
            engine = self.inputs.config.create_engine()
            metadata_obj = sqla.MetaData()
            table = sqla.Table("uvvsmodels", metadata_obj, autoload_with=engine)
            
            query = sqla.select(table).where(
                table.columns.out_idnum == oid,
                table.columns.quantity == self.quantity,
                table.columns.query == self.query,
                table.columns.dphi == self.dphi,
                table.columns.mechanism == self.mechanism,
                table.columns.wavelength == [w.value for w in self.wavelength],
                table.columns.fitted == self.fitted)
                
            with engine.connect() as con:
                result = pd.DataFrame(con.execute(query))
                
            # Should only have one match per outputfile
            assert len(result) <= 1
            
            if len(result) == 0:
                search_results[outputfile] = None
            else:
                # from IPython import embed; embed()
                # import sys; sys.exit()
                search_results[outputfile] = (result.loc[0, 'idnum'],
                                              result.loc[0, 'unfit_idnum'],
                                              result.loc[0, 'filename'])
        
        return search_results
    
    def restore(self, search_result):
        # Restore is on an outputfile basis
        idnum, ufit_idnum, modelfile = search_result
        print(f'Restoring modelfile {modelfile}.')
        with open(modelfile, 'rb') as f:
            iteration_result = pickle.load(f)
        
        iteration_result.modelfile = modelfile
        iteration_result.model_idnum = idnum
        if isinstance(iteration_result, FittedIterationResult):
            self.ufit_idnum = ufit_idnum
        else:
            pass
        
        return iteration_result
    
    def _data_setup(self, data):
        # distance of s/c from planet
        dist_from_plan = np.sqrt(data.x**2 + data.y**2 + data.z**2)
        
        # Angle between look direction and planet.
        ang = np.arccos((-data.x * data.xbore - data.y * data.ybore -
                         data.z * data.zbore) / dist_from_plan)
        
        # Check to see if look direction intersects the planet anywhere
        asize_plan = np.arcsin(1. / dist_from_plan)
        
        # Don't worry about lines of sight that don't hit the planet
        dist_from_plan.loc[ang > asize_plan] = 1e30
        
        return dist_from_plan
    
    def _spectrum_process(self, spectrum, packets, tree, dist,
                          i=None, find_weighting=False, weight_info=None):
        x_sc = spectrum[xcols].values.astype(float)
        bore = spectrum[borecols].values.astype(float)
        
        dd = 30  # Furthest distance we need to look
        x_far = x_sc + bore * dd
        while np.linalg.norm(x_far) > self.oedge:
            dd -= 0.1
            x_far = x_sc + bore * dd
        
        t = [0.05]
        while t[-1] < dd:
            t.append(t[-1] + t[-1] * np.sin(self.dphi))
        t = np.array(t)
        Xbore = x_sc[np.newaxis, :] + bore[np.newaxis, :] * t[:, np.newaxis]
        
        wid = t * np.sin(self.dphi)
        ind = np.concatenate(tree.query_radius(Xbore, wid))
        ilocs = np.unique(ind).astype(int)
        indicies = packets.iloc[ilocs].index
        subset = packets.loc[indicies]
        
        xpr = subset[xcols] - x_sc[np.newaxis, :]
        rpr = np.sqrt(xpr['x'] * xpr['x'] +
                      xpr['y'] * xpr['y'] +
                      xpr['z'] * xpr['z'])
        
        losrad = np.sum(xpr * bore[np.newaxis, :], axis=1)
        inview = rpr < dist
        
        if np.any(inview):
            # used_packets = inview.index
            # used_packets0 = packets.Index[inview.index].unique()
            
            Apix = np.pi * (rpr[inview] * np.sin(self.dphi))**2 * (
                self.unit.to(u.cm))**2
            wtemp = subset.loc[inview, 'weight'] / Apix
            if self.quantity == 'radiance':
                # Determine if any packets are in shadow
                # Projection of packet onto LOS
                # Point along LOS the packet represents
                losrad_ = losrad[inview].values
                hit = (x_sc[np.newaxis, :] +
                       bore[np.newaxis, :] * losrad_[:, np.newaxis])
                rhohit = np.linalg.norm(hit[:, [0, 2]], axis=1)
                out_of_shadow = (rhohit > 1) | (hit[:, 1] < 0)
                wtemp *= out_of_shadow
                
                rad = wtemp.sum()
                pack = np.sum(inview)
                
                if (rad > 0) and find_weighting:
                    ratio = spectrum.radiance / rad
                    
                    # Save which packets are used for each spectrum
                    weight_info['saved_packets'].loc[i] = subset.loc[inview, 'Index'].unique()
                    
                    should = weight_info['weighting'].index.to_series().apply(
                        self._should_add_weight, args=(weight_info['saved_packets'].loc[i],))
                    
                    weight_info['weighting'].loc[should] = (
                        weight_info['weighting'].loc[should].apply(self._add_weight,
                                                                   args=(ratio,)))
                    weight_info['included'].loc[should] = (
                        weight_info['included'].loc[should].apply(self._add_index,
                                                                  args=(i,)))
                else:
                    pass
            else:
                assert False, 'Other quantities not set up.'
        else:
            rad, pack, = 0., 0
            
        return rad, pack
    
    def _tree(self, values, type='KDTree'):
        if type == 'KDTree':
            return KDTree(values)
        elif type == 'BallTree':
            return BallTree(values)
        
    def make_mask(self, data):
        mask = np.array([True for _ in data.radiance])
        sigmalimit = None
        if self.masking is not None:
            for masktype in self.masking.split(';'):
                masktype = masktype.strip().lower()
                if masktype.startswith('middle'):
                    perinterval = float(masktype[6:])
                    # Estimate model strength (source rate) by fitting middle %
                    interval = PercentileInterval(perinterval)
                    lim = interval.get_limits(data)
                    mask = (mask &
                            (data.radiance >= lim[0]) &
                            (data.radiance <= lim[1]))
                elif masktype.startswith('minalt'):
                    minalt = float(masktype[6:])
                    mask = mask & (data.alttan >= minalt)
                elif masktype.startswith('minsnr'):
                    minSNR = float(masktype[6:])
                    snr = data.radiance/data.sigma
                    mask = mask & (snr > minSNR)
                elif masktype.startswith('siglimit'):
                    sigmalimit = float(masktype[8:])
                else:
                    raise ValueError('nexoclom.math.fit_model',
                                     f'masking = {masktype} not defined.')
        else:
            pass
        
        return mask, sigmalimit

    def determine_source_from_data(self, scdata):
        # Search for unfitted outputfiles
        self.fitted = True
        self.inputs.options.fitted = False
        unfit_outid, unfit_outputfiles, unfit_npackets, _ = self.inputs.search()
        if unfit_npackets == 0:
            raise RuntimeError('No packets found for these Inputs.')
        else:
            self.unfit_outid, self.unfit_outputfiles = unfit_outid, unfit_outputfiles
            self.inputs.options.fitted = True
            
            self.inputs.spatialdist = SpatialDist({'type': 'fitted output'})
            self.inputs.spatialdist.query = scdata.query
            
            self.inputs.speeddist= SpeedDist({'type': 'fitted output'})
            self.inputs.speeddist.query = scdata.query

        data = scdata.data
        iteration_results = []
        dist_from_plan = self._data_setup(data)

        # Determine which points should be used for the fit
        mask, _ = self.make_mask(data)

        for ufit_id, ufit_out in zip(self.unfit_outid, self.unfit_outputfiles):
            # Search for fitted outputfiles
            self.inputs.spatialdist.unfit_outid = ufit_id
            self.inputs.speeddist.unfit_outid = ufit_id
            self.outid, self.outputfiles, _, _ = self.inputs.search()
            assert len(self.outid) <= 1
            
            # Search for completed fitted models
            search_results = self.search()
            if len(self.outid) == 1:
                search_result = search_results.get(self.outputfiles[0], None)
            else:
                search_result = None
            
            if search_result is None:
                output = Output.restore(ufit_out)
                
                packets = copy.deepcopy(output.X)
                packets['radvel_sun'] = (packets['vy'] +
                                         output.vrplanet.to(self.unit / u.s).value)

                self.oedge = output.inputs.options.outeredge * 2
                
                # Will base shadow on line of sight, not the packets
                out_of_shadow = np.ones(packets.shape[0])
                self.packet_weighting(packets, out_of_shadow, output.aplanet)
                
                # This sets limits on regions where packets might be
                tree = self._tree(packets[xcols].values)
                
                # rad = modeled radiance
                # weighting = list of the weights that should be applied
                #   - Final weighting for each packet is mean of weights
                rad = pd.Series(np.zeros(data.shape[0]), index=data.index)
                ind0 = packets.Index.unique()
                weight_info = {
                    'saved_packets': pd.Series((np.ndarray((0,), dtype=int)
                                                for _ in range(data.shape[0])),
                                               index=data.index),
                   'weighting': pd.Series((np.ndarray((0,))
                                           for _ in range(ind0.shape[0])),
                                          index=ind0),
                   'included': pd.Series((np.ndarray((0,), dtype=np.int)
                                          for _ in range(ind0.shape[0])),
                                          index=ind0)}
                    
                print(f'{data.shape[0]} spectra taken.')
                for i, spectrum in data.iterrows():
                    rad_, _ = self._spectrum_process(spectrum, packets, tree,
                                                     dist_from_plan[i],
                                                     find_weighting=mask[i],
                                                     i=i, weight_info=weight_info)
                    rad.loc[i] = rad_
                    
                    if len(data) > 10:
                        ind = data.index.get_loc(i)
                        if (ind % (len(data) // 10)) == 0:
                            print(f'Completed {ind + 1} spectra')
                    
                assert np.all(weight_info['weighting'].apply(len) ==
                              weight_info['included'].apply(len))

                # Determine the proper weightings
                new_weight = weight_info['weighting'].apply(
                    lambda x:x.mean() if x.shape[0] > 0 else 0.)
                new_weight /= new_weight[new_weight > 0].mean()
                assert np.all(np.isfinite(new_weight))
                
                if np.any(new_weight > 0):
                    multiplier = new_weight.loc[output.X['Index']].values
                    output.X.loc[:, 'frac'] = output.X.loc[:, 'frac'] * multiplier
                    output.X0.loc[:, 'frac'] = output.X0.loc[:, 'frac'] * new_weight
                    
                    output.X = output.X[output.X.frac > 0]
                    output.X0 = output.X0[output.X0.frac > 0]
                    output.totalsource = output.X0['frac'].sum() * output.nsteps
                    
                    # Save the fitted output
                    output.inputs = self.inputs
                    output.save()
                    
                    # Find the radiance again with the new output
                    packets = copy.deepcopy(output.X)
                    packets['radvel_sun'] = (packets['vy'] +
                                             output.vrplanet.to(self.unit / u.s).value)
                    out_of_shadow = np.ones(packets.shape[0])
                    self.packet_weighting(packets, out_of_shadow, output.aplanet)
                    tree = self._tree(packets[xcols].values)
                    rad = pd.Series(np.zeros(data.shape[0]), index=data.index)

                    for i, spectrum in data.iterrows():
                        rad_, _ = self._spectrum_process(spectrum, packets, tree,
                                                         dist_from_plan[i],
                                                         find_weighting=False,
                                                         i=i)
                        rad.loc[i] = rad_
    
                        if len(data) > 10:
                            ind = data.index.get_loc(i)
                            if (ind % (len(data) // 10)) == 0:
                                print(f'Completed {ind + 1} spectra')

                    # Save the starting state for making a source map
                    # longitude = np.append(longitude, output.X0.longitude)
                    # latitude = np.append(latitude, output.X0.latitude)
                    # vel_ = np.sqrt(output.X0.vx**2 + output.X0.vy**2 +
                    #                output.X0.vz**2) * self.inputs.geometry.planet.radius
                    # velocity = np.append(velocity, vel_)
                    # weight = np.append(weight, output.X0.frac)
                    
                    iteration = {'radiance': rad,
                                 'npackets': output.X0.frac.sum(),
                                 'totalsource': output.totalsource,
                                 'outputfile': output.filename,
                                 'out_idnum': output.idnum,
                                 'unfit_outputfile': ufit_out,
                                 'unfit_outid': ufit_id}
                    iteration_result = FittedIterationResult(iteration)
                    iteration_result.saved_packets = weight_info['saved_packets']
                    iteration_result.weighting = weight_info['weighting']
                    iteration_result.included = weight_info['included']
                else:
                    iteration = {'radiance': rad,
                                 'npackets': 0.,
                                 'totalsource': 0.,
                                 'outputfile': output.filename,
                                 'out_idnum': output.idnum,
                                 'unfit_outputfile': ufit_out,
                                 'unfit_outid': ufit_id}
                    iteration_result = FittedIterationResult(iteration)
                    iteration_result.saved_packets = weight_info['saved_packets']
                    iteration_result.weighting = weight_info['weighting']
                    iteration_result.included = weight_info['included']

                modelfile = self.save(iteration_result)
                iteration_result.modelfile = modelfile
                iteration_results.append(iteration_result)
                del output
            else:
                # Restore saved result
                print(f'Using saved file {search_result[2]}')
                iteration_result = self.restore(search_result)
                assert len(iteration_result.radiance) == len(data)
                iteration_result.model_idnum = search_result[0]
                iteration_result.modelfile = search_result[2]
                iteration_results.append(iteration_result)

        # Combine iteration_results into single new result
        self.modelfiles = {}
        for iteration_result in iteration_results:
            self.radiance += iteration_result.radiance
            self.totalsource += iteration_result.totalsource
            self.modelfiles[iteration_result.outputfile] = iteration_result.modelfile
        
        self.outputfiles = self.modelfiles.keys()
        model_rate = self.totalsource/self.inputs.options.endtime.value
        self.atoms_per_packet = 1e23 / model_rate
        self.radiance *= self.atoms_per_packet/1e3*u.kR
        self.determine_source_rate(scdata)
        self.atoms_per_packet *= self.sourcerate.unit
        self.outputfiles = list(self.modelfiles.keys())

        print(self.totalsource, self.atoms_per_packet)
    
    def simulate_data_from_inputs(self, scdata):
        """Given a set of inputs, determine what the spacecraft should see.
        Models should have already been run.
        
        **Outputs**
        """
        # If using a planet-fixed source map, need to set subsolarlon
        if ((self.inputs.spatialdist.type == 'surface map') and
            (self.inputs.spatialdist.coordinate_system == 'planet-fixed')):
            self.inputs.spatialdist.subsolarlon = scdata.subslong.median() * u.rad
        else:
            pass
        
        # This is will work with fitted or non-fitted outputfiles
        self.outid, self.outputfiles, self.npackets, self.totalsource = self.inputs.search()
        if self.npackets == 0:
            raise RuntimeError('No packets found for these Inputs.')

        data = scdata.data
        search_results = self.search()
        iteration_results = []
        
        # Do this step if will need to compute any iteration results
        dist_from_plan = (self._data_setup(data)
                          if None in search_results.values()
                          else None)
        for outputfile, search_result in search_results.items():
            if search_result is None:
                # simulate the data
                output = Output.restore(outputfile)
                
                packets = copy.deepcopy(output.X)
                packets['radvel_sun'] = (packets['vy'] +
                                         output.vrplanet.to(self.unit / u.s).value)
                self.oedge = output.inputs.options.outeredge * 2
                
                # Will base shadow on line of sight, not the packets
                out_of_shadow = np.ones(len(packets))
                self.packet_weighting(packets, out_of_shadow, output.aplanet)
                
                # This sets limits on regions where packets might be
                tree = self._tree(packets[xcols].values)
                
                rad = pd.Series(np.zeros(data.shape[0]), index=data.index)
                npack = pd.Series(np.zeros(data.shape[0]), index=data.index,
                                  dtype=int)
                print(f'{data.shape[0]} spectra taken.')
                for i, spectrum in data.iterrows():
                    rad_, pack_ = self._spectrum_process(spectrum, packets, tree,
                                                         dist_from_plan[i])
                    rad.loc[i] = rad_
                    npack.loc[i] = pack_
                    
                    if len(data) > 10:
                        ind = data.index.get_loc(i)
                        if (ind % (len(data) // 10)) == 0:
                            print(f'Completed {ind + 1} spectra')
                
                iteration_ = {'radiance': rad,
                              'npackets': npack,
                              'totalsource': output.totalsource,
                              'outputfile': outputfile,
                              'out_idnum': output.idnum,
                              'query': scdata.query}
                iteration_result = IterationResult(iteration_)
                modelfile = self.save(iteration_result)
                iteration_result.modelfile = modelfile
                iteration_results.append(iteration_result)
            else:
                print(f'Using saved result {search_result[2]}')
                iteration_result = self.restore(search_result)
                iteration_result.model_idnum = search_result[0]
                iteration_result.modelfile = search_result[2]
                assert len(iteration_result.radiance) == len(data)
                iteration_results.append(iteration_result)

        # combine iteration_results
        self.modelfiles = {}
        for iteration_result in iteration_results:
            self.radiance += iteration_result.radiance
            self.modelfiles[iteration_result.outputfile] = iteration_result.modelfile
            
        # need model rate for this output
        model_rate = self.totalsource / self.inputs.options.endtime.value
        self.atoms_per_packet = 1e23 / model_rate
        self.radiance *= self.atoms_per_packet/1e3  # kR
        self.determine_source_rate(scdata)
        self.atoms_per_packet *= self.sourcerate.unit
        self.outputfiles = self.modelfiles.keys()
        
        print(self.totalsource, self.atoms_per_packet)
        
        # self.sourcemap = self.make_source_map(longitude, latitude, velocity,
        #                                       weight)
        
    def determine_source_rate(self, scdata):
        mask, sigmalimit = self.make_mask(scdata.data)
        linmodel = models.Multiply()
        fitter = fitting.LinearLSQFitter()
        best_fit = fitter(linmodel, self.radiance.values[mask],
                          scdata.data.radiance.values[mask],
                          weights=1./scdata.data.sigma.values[mask]**2)
        
        if sigmalimit is not None:
            diff = np.abs((scdata.data.radiance.values -
                           best_fit.factor*self.radiance.values)/
                          scdata.data.sigma)
            mask = mask & (diff < sigmalimit)
            best_fit = fitter(linmodel, self.radiance.values[mask],
                              scdata.data.radiance.values[mask],
                              weights=1./scdata.data.sigma.values[mask]**2)
        else:
            pass


        self.radiance *= best_fit.factor.value
        self.sourcerate = best_fit.factor.value * u.def_unit('10**23 atoms/s', 1e23 / u.s)
        self.goodness_of_fit = None
        self.mask = mask



"""This loads the generic solar system kernels

Kernels to load:
    * leap second
    * spk for solar system and moons
    * planetary constants
    * frames if available
"""
import spiceypy as spice
import os
import glob

path = '/Users/mburger/Work/Data/spice_kernels/Generic'

# Need to do better than hard code it
def load_kernels():
    ## Determine the path to the spice kernels
    kernels = []

    ## Load the leapseconds (lsk kernel)
    lsk_kernel = glob.glob(os.path.join(path, 'lsk', '*.tls'))
    lsk_kernel = lsk_kernel[-1]
    spice.furnsh(lsk_kernel)
    kernels.append(lsk_kernel)

    ##Generic Solar System ephemeris kernels
    ##  1) load all the planet and satellite kernels
    ##      Note 11/20/2014:
    ##      DE430 is the appropriate planetary kernel for all the planets except Pluto
    ##      DE432 is provided for Pluto

    spk_kernel = ['spk/planets/de430.bsp',
    #  'spk/satellites/mar097.bsp',
      'spk/satellites/jup310.bsp'
    #  'spk/satellites/sat365.bsp',
    #  'spk/satellites/ura111.bsp',
    #  'spk/satellites/nep081.bsp',
    #  'spk/satellites/nep077.bsp',
    #  'spk/satellites/plu043.bsp',
     ]

    for kern in spk_kernel:
        spice.furnsh(os.path.join(path, kern))
        kernels.append(path+kern)

    ## Load the planetary constants (pck kernel)
    pck_kernel = os.path.join(path, 'pck/pck00010.tpc')
    spice.furnsh(pck_kernel)
    kernels.append(pck_kernel)

    ## Load the frames kernels if available
    fk_kernel = ['JEF.tk', 'SEF.tf', 'SMF.tf']
    for kern in fk_kernel:
        kk = os.path.join(path, 'Homemade/fk/', kern)
        if os.path.isfile(kk):
            spice.furnsh(kk)
            kernels.append(kk)

    return kernels



"""Create and read configuration file, create necessary database tables."""
import os
import pandas as pd
from sqlalchemy import text
from nexoclom.utilities import NexoclomConfig
from nexoclom import __file__ as basefile


basepath = os.path.dirname(basefile)


def configure_nexoclom(verbose=False):
    # Create the database if necessary
    config = NexoclomConfig(verbose=verbose)
    config.verify_database_running()
    
    # Validate nexoclom output tables
    with config.create_engine().begin() as con:
        with open(os.path.join(basepath, 'data', 'schema.sql'), 'r') as sqlfile:
            done = False
            while not done:
                line = sqlfile.readline()
                nextline = ''
                if 'CREATE' in line:
                    if 'TYPE' in line:
                        name = line.split(' ')[2]
                    elif 'TABLE' in line:
                        name = line.split(' ')[5]
                    else:
                        assert False
                    
                    exists = pd.DataFrame(con.execute(text(
                        f'''SELECT exists (SELECT 1 FROM pg_type
                                           WHERE typname = :name);'''),
                        {'name': name}))
                    
                    if not exists.loc[0, 'exists']:
                        # Create the table if it isn't there
                        query = line
                        nextline = sqlfile.readline()
                        while (nextline.strip()) and ('DONE' not in nextline):
                            query += nextline
                            nextline = sqlfile.readline()
                            
                        print(query)
                        con.execute(text(query))
                    else:
                        pass
                else:
                    pass
                
                done = ('DONE' in nextline) or ('DONE' in line)
    return config

def configure_solarsystem():
    # Make a pickle file with the planetary constants
    pklfile = os.path.join(basepath, 'data', 'PlanetaryConstants.pkl')
    if not os.path.exists(pklfile):
        constfile = os.path.join(basepath, 'data', 'PlanetaryConstants.dat')
        constants = pd.read_csv(constfile, sep=':', comment='#')
        constants.columns = [col.strip() for col in constants.columns]
        constants['Object'] = constants['Object'].apply(lambda x: x.strip())
        constants['orbits'] = constants['orbits'].apply(lambda x: x.strip())
        constants.to_pickle(pklfile)
    else:
        pass

    # Do something with the naif_ids.
    # pklfile = os.path.join(basepath, 'data', 'naif_ids.pkl')
    # if not os.path.exists(pklfile):
        # naiffile = os.path.join(basepath, 'naif_ids.dat')

def configure_atomicdata():
    # Make gvalue table
    gvalue_file = os.path.join(basepath, 'data', 'g-values', 'g-values.pkl')
    if not os.path.exists(gvalue_file):
        from nexoclom.atomicdata.initialize_atomicdata import make_gvalue_table
        make_gvalue_table()
    else:
        pass

    # Make the photorates table
    photorates_file = os.path.join(basepath, 'data', 'Loss', 'photorates.pkl')
    if not os.path.exists(photorates_file):
        from nexoclom.atomicdata.initialize_atomicdata import make_photorates_table
        make_photorates_table()
        
def configure(verbose=False):
    config = configure_nexoclom(verbose=True)
    configure_solarsystem()
    configure_atomicdata()
    
    return config



import numpy as np
from sklearn.neighbors import KDTree, BallTree


class Histogram:
    """ Wrapper for np.histogram that makes the x-axis the center of each bin.
    Returns a class with everything self-contained.
    """
    def __init__(self, a, bins=10, range=None, normed=None, weights=None,
                 density=None):
        hist, x = np.histogram(a, bins, range, normed, weights, density)
        self.histogram = hist.astype(np.float)
        self.dx = x[1]-x[0]  # width of the bin
        self.x = x[:-1] + self.dx/2
        
        
class Histogram2d:
    """ Wrapper for np.histogram2d that makes the x,y axes the centers of each bin.
    Returns a class with everything self-contained.
    """
    def __init__(self, ptsx, ptsy, bins=10, range=None, normed=None, weights=None,
                 density=None):
        hist, x, y = np.histogram2d(ptsx, ptsy, bins, range, normed, weights, density)
        self.histogram = hist
        self.dx, self.dy = x[1]-x[0], y[1]-y[0]
        self.x = x[:-1] + self.dx/2
        self.y = y[:-1] + self.dy/2


# class HistogramSphere:
#     def __init__(self, longitude, latitude, weights=None,
#                  dphi=np.deg2rad(5), bins=(180, 90)):
#         ptsx = np.cos(longitude) * np.cos(latitude)
#         ptsy = np.sin(longitude) * np.cos(latitude)
#         ptsz = np.sin(latitude)
#
#         if isinstance(bins, tuple) or isinstance(bins, list):
#             self.bins = bins
#         elif isinstance(bins, int):
#             self.bins = (bins, bins)
#         else:
#             raise TypeError('bins must be an int, tuple, or list')
#
#         self.dphi = dphi
#
#         self.longitude = np.linspace(0, 2*np.pi, bins[0], endpoint=False)
#         self.longitude += (self.longitude[1]-self.longitude[0])/2.
#         self.latitude = np.linspace(-np.pi/2, np.pi/2, bins[1], endpoint=False)
#         self.latitude += (self.latitude[1]-self.latitude[0])/2.
#
#         gridlongitude, gridlatitude = np.meshgrid(self.longitude, self.latitude)
#         gridx = (np.cos(gridlongitude) * np.cos(gridlatitude)).flatten()
#         gridy = (np.sin(gridlongitude) * np.cos(gridlatitude)).flatten()
#         gridz = np.sin(gridlatitude).flatten()
#         grid = np.array([gridx, gridy, gridz]).T
#
#         tree = KDTree(np.array([ptsx, ptsy, ptsz]).T)
#         ind = tree.query_radius(grid, self.dphi)
#         if weights is None:
#             result = np.array([len(x) for x in ind])
#         else:
#             result = np.array([np.sum(weights[x]) for x in ind])
#
#         self.histogram = result.reshape(gridlatitude.shape)


''' Determine some basic geometry info from the SPICE kernels

For planets:
  * r in AU
  * drdt in km/s
  * TAA in radians
  * Sub-solar longitude and latitude in radians  -- not implemented
  * Sub-earth longitude and latitude in radians  -- not implemented
'''
import numpy as np
import astropy.units as u
import astropy.constants as const
from nexoclom.solarsystem.relative_position import relative_position
from nexoclom.solarsystem.SSObject import SSObject

def planet_geometry(time, planet):
    ## Determine planet location relative to sun
    # if planet.lower() == 'mercury':
    #     frame0 = 'MSGR_MSO'
    #     frame1 = 'IAU_MERCURY'
    # elif planet.lower() == 'jupiter':
    #     frame0 = 'Jupiter_Equatorial_Frame'
    #     frame1 = 'IAU_JUPITER'
    # elif planet.lower() == 'saturn':
    #     frame0 = 'Saturn_Equatorial_Frame'
    #     frame1 = 'IAU_SATURN'
    # else:
    #     print('Proper refrerence frames not defined for %s.' % planet)
    #     return -1

    # x0, v0 = relative_position(planet, 'Sun', time, frame=frame0)
    # x1, v1 = relative_position(planet, 'Sun', time, frame=frame1)
    x2, v2 = relative_position(planet, 'Sun', time, frame='J2000')
    #
    # e0, ve0 = relative_position(planet, 'Earth', time, frame=frame0)
    # e1, ve1 = relative_position(planet, 'Earth', time, frame=frame1)

    ## Compute TAA
    mu = const.M_sun * const.G
    x2, v2, mu = x2.value, v2.value, mu.to(u.km**3/u.s**2).value

    ## r and drdt from sun
    r2 = np.sqrt(np.sum(x2**2))
    drdt = np.sum(x2*v2)/r2

    ## Eccentricity vector
    e = np.sum(v2**2)*x2/mu - np.sum(x2*v2)*v2/mu - x2/r2
    ee = np.sqrt(np.sum(e**2))
    taa = np.arccos(np.sum(e*x2)/ee/r2)
    if np.sum(x2*v2) < 0:
        taa = 2*np.pi - taa

    return {'r':r2*u.km, 'drdt':drdt*u.km/u.s, 'taa':taa*u.rad}



""" Tests atomicdata.atomicmass
Tests
    Three random elements
    H2O, CO2, C2G2OH
    X (non existing element)
    
Compares with masses in periodictable (i.e., if periodictable changes, tests
will still succeed.
"""
import astropy.units as u
import periodictable as pt
import random
from nexoclom.atomicdata import atomicmass
import pytest


sp = [random.choice(list(pt.elements)).symbol for _ in range(3)]
sp.extend(['H2O', 'CO2', 'C2H2OH'])
res = [pt.formula(s).mass for s in sp]
sp.append('X')
res.append(None)

args = [pytest.param(s, r, id=s) for s,r in zip(sp, res)]

@pytest.mark.atomicdata
@pytest.mark.parametrize('species, result', args)
def test_atomicmass(species, result):
    if result is None:
        assert atomicmass(species) is None
    else:
        assert atomicmass(species) == result * u.u, f'{species} mass failure'



from nexoclom.solarsystem import SSObject
import pytest


@pytest.mark.solarsystem
def test_SSObject():
    p0 = SSObject('Mercury')
    assert p0.object == 'Mercury'
    assert p0.type == 'Planet'
    assert p0.moons is None
    assert len(p0) == 1

    p1 = SSObject('Jupiter')
    assert p1.object == 'Jupiter'
    assert p1.type == 'Planet'
    assert len(p1.moons) == 4
    assert len(p1) == 5
    
    p2 = SSObject('Io')
    assert p2.object == 'Io'
    assert p2.type == 'Moon'
    assert p2.orbits == 'Jupiter'
    
    p3 = SSObject('Fake')
    assert p3.object is None
    
    p4 = SSObject('Sun')
    assert p4.object == 'Sun'
    assert p4.type == 'Star'
    assert p4.orbits == 'Milky Way'

    p5 = SSObject('Sun')
    assert p4 == p5
    assert p1 != p5



"""Version of LOSResult_pd1.py using dask to handle large dataframes"""
import os.path
import numpy as np
import pandas as pd
import dask.dataframe as dd
import pickle
import random
import copy
import astropy.units as u
from sklearn.neighbors import KDTree, BallTree
from scipy.interpolate import RectBivariateSpline
import scipy.linalg as lin
from nexoclom.math import fit_model

from nexoclom.modelcode import ModelResult
from nexoclom.utilities import database_connect
from nexoclom import Input
from nexoclom import Output


xcols = ['x', 'y', 'z']
borecols = ['xbore', 'ybore', 'zbore']
NLONBINS, NLATBINS, NVELBINS = 72, 36, 100


class InputError(Exception):
    """Raised when a required parameter is not included."""
    def __init__(self, expression, message):
        self.expression = expression
        self.message = message


class LOSResult(ModelResult):
    """Class to contain the LOS result from multiple outputfiles."""
    def __init__(self, scdata, quantity='radiance', dphi=1*u.deg):
        """Determine column or emission along lines of sight.
        This assumes the model has already been run.
        
        Parameters
        ==========
        scdata
            Spacecraft data object (currently designed for MESSENGERdata object
            but can be faked for other types of data)
            
        quantity
            Quantity to calculate: 'column', 'radiance', 'density'
            
        dphi
            Angular size of the view cone. Default = r deg.
        """
        format_ = {'quantity':quantity}
        super().__init__(format_, species=scdata.species)
        
        # Basic information
        self.scdata = scdata
        self.scdata.set_frame('Model')
        self.type = 'LineOfSight'
        self.unit = None
        self.dphi = dphi.to(u.rad).value
        self.oedge = None
        
        self.fitted = None
        self.outputfiles = []
        nspec = len(self.scdata)
        self.radiance = pd.Series(np.zeros(nspec), index=scdata.data.index)
        self.npackets = pd.Series(np.zeros(nspec), index=scdata.data.index,
                                  dtype=int)
        self.totalsource = 0.
        self.sourcemap = None
        self.modelfiles = {}
        self.surfacefiles = {}
        self.data_packets = {'datafiles': [],
                             'totalsource': 0.}

    def delete_models(self):
        """Deletes any LOSResult models associated with this data and input
        This may never actually do anything. Overwrite=True will also
        erase the outputfiles (which erases any models that depend on them).
        Unless I put separate outputfile and modelfile delete switches,
        This shouldn't do anything"""
        
        search_results = self.search()
        if len(search_results) != 0:
            print('Warning: LOSResult.delete_models found something to delete')
            for _, search_result in search_results.items():
                if search_result is not None:
                    idnum, modelfile = search_result
                    with database_connect() as con:
                        cur = con.cursor()
                        cur.execute(f'''DELETE from uvvsmodels
                                       WHERE idnum = %s''', (idnum, ))
                    if os.path.exists(modelfile):
                        os.remove(modelfile)
        else:
            pass
        
    def surfacefile_name(self, modelfile):
        surf_file = os.path.join(os.path.dirname(modelfile),
                                 os.path.basename(modelfile).replace(
                                     'model', 'surface').replace('.pkl', '.parquet'))
        return surf_file

    def save(self, data_packets, surface):
        # Insert the model into the database
        # Save is on an outputfile basis
        if self.quantity == 'radiance':
            mech = ', '.join(sorted([m for m in self.mechanism]))
            wave_ = sorted([w.value for w in self.wavelength])
            wave = ', '.join([str(w) for w in wave_])
        else:
            mech = None
            wave = None

        # Save query with all white space removed and lowercase
        tempname = f'temp_{str(random.randint(0, 1000000))}'

        with database_connect() as con:
            cur = con.cursor()
            cur.execute(f'''INSERT into uvvsmodels (out_idnum, quantity,
                            query, dphi, mechanism, wavelength,
                            fitted, filename)
                            values (%s, %s, %s, %s, %s, %s, %s, %s)''',
                        (data_packets['out_idnum'], self.quantity,
                         self.scdata.query, self.dphi, mech, wave, self.fitted,
                         tempname))

            # Determine the savefile name
            idnum_ = pd.read_sql(f'''SELECT idnum
                                     FROM uvvsmodels
                                     WHERE filename='{tempname}';''', con)
            assert len(idnum_) == 1
            idnum = int(idnum_.idnum[0])

            savefile = os.path.join(os.path.dirname(data_packets['outputfile']),
                                    f'model.{idnum}.pkl')

            cur.execute(f'''UPDATE uvvsmodels
                            SET filename=%s
                            WHERE idnum=%s''', (savefile, idnum))
            
        with open(savefile, 'wb') as f:
            pickle.dump(data_packets, f)
        if surface is None:
            savefile_surf = None
        else:
            savefile_surf = self.surfacefile_name(savefile)
            surface.to_parquet(savefile_surf)
        
        return savefile, savefile_surf

    def search(self):
        """
        :return: dictionary containing search results:
                 {outputfilename: (modelfile_id, modelfile_name)}
        """
        search_results = {}
        for outputfile in self.outputfiles:
            with database_connect() as con:
                # Determine the id of the outputfile
                idnum_ = pd.read_sql(
                    f'''SELECT idnum
                        FROM outputfile
                        WHERE filename='{outputfile}' ''', con)
                oid = idnum_.idnum[0]
            
                if self.quantity == 'radiance':
                    mech = ("mechanism = '" +
                            ", ".join(sorted([m for m in self.mechanism])) +
                            "'")
                    wave_ = sorted([w.value for w in self.wavelength])
                    wave = ("wavelength = '" +
                            ", ".join([str(w) for w in wave_]) +
                            "'")
                else:
                    mech = 'mechanism is NULL'
                    wave = 'wavelength is NULL'
            
                result = pd.read_sql(
                    f'''SELECT idnum, filename FROM uvvsmodels
                        WHERE out_idnum={oid} and
                              quantity = '{self.quantity}' and
                              query = '{self.scdata.query}' and
                              dphi = {self.dphi} and
                              {mech} and
                              {wave} and
                              fitted = {self.fitted}''', con)
            
                # Should only have one match per outputfile
                assert len(result) <= 1
                
                if len(result) == 0:
                    search_results[outputfile] = None
                else:
                    surffile = self.surfacefile_name(result.iloc[0,1])
                    search_results[outputfile] = (result.iloc[0, 0],
                                                  result.iloc[0, 1],
                                                  surffile)
                
        return search_results
    
    @staticmethod
    def restore(modelfile):
        # Restore is on an outputfile basis
        with open(modelfile, 'rb') as f:
            data_packets = pickle.load(f)

        return data_packets
    
    def _data_setup(self):
        # distance of s/c from planet
        data = self.scdata.data
        dist_from_plan = np.sqrt(data.x**2 + data.y**2 + data.z**2)
    
        # Angle between look direction and planet.
        ang = np.arccos((-data.x*data.xbore - data.y*data.ybore -
                         data.z*data.zbore)/dist_from_plan)
        
        # Check to see if look direction intersects the planet anywhere
        asize_plan = np.arcsin(1./dist_from_plan)

        # Don't worry about lines of sight that don't hit the planet
        dist_from_plan.loc[ang > asize_plan] = 1e30
        
        return dist_from_plan
    
    @staticmethod
    def _tree(values, treetype='KDTree'):
        if treetype == 'KDTree':
            return KDTree(values)
        elif treetype == 'BallTree':
            return BallTree(values)
        
    def determine_source_from_data(self, modnum, weight_method='scaling'):
        modkey = f'model{modnum:02d}'
        maskkey = f'mask{modnum:02d}'
        mask = self.scdata.data[maskkey]
        
        self.modelfiles = self.scdata.model_info[modkey]['modelfiles']
        assert len(self.modelfiles) > 0
        self.outputfiles = self.modelfiles.keys()
        data = self.scdata.data
        totalsource, nsteps, endtime = 0, 0, 0
        datafiles, surfacefiles = [], []
        for outputfile, modelfile in self.modelfiles.items():
            # Restore the unfit output file
            data_packets_it = self.restore(modelfile)
            
            print('Preparing dask dataframe')
            datafiles.extend(data_packets_it['datafiles'])
            surfacefiles.extend(self.surfacefile_name(modelfile))
            totalsource += data_packets_it['totalsource']
            endtime = data_packets_it['endtime']
            nsteps = data_packets_it['nsteps']
            
        if weight_method == 'scaling':
            # Determine the proper weightings
            data_df = dd.read_parquet(datafiles)
            mask_ = mask[mask].index
            data_df = data_df[data_df.specind.isin(mask_)]
            
            spec_ind = data_df.set_index('specind')
            scale_factor_ = spec_ind.groupby(['outputfile', 'Index0'])['ratio'].mean()
            scale_factor_ /= scale_factor_[scale_factor_ > 0].mean()
            
            data_df['scale_factor'] = np.zeros(len(self.data_packets.data))
            ind0_ind = self.data_packets.data.set_index(['outputfile', 'Index0'])
            ind0_ind.loc[scale_factor_.index, 'scale_factor'] = scale_factor_
            self.data_packets.data = ind0_ind.reset_index()
            self.data_packets.data['fit_weight'] = (self.data_packets.data.weight *
                                                    self.data_packets.data.scale_factor)
            from IPython import embed; embed()
            import sys; sys.exit()
    
            # Determine surface weighting
            surface_df = dd.read_parquet(surfacefiles)
            surface_ind = surface_source.set_index(['outputfile', 'Index'])
            surface_ind.loc[scale_factor_.index, 'weight'] = scale_factor_
            surface_ind.dropna(inplace=True)
            surface_source = surface_ind.reset_index()
        # elif weight_method == 'leastsq':
        #     # This method produces too many scale_factors < 0
        #     assert 0
        #
        #     # Solve linear equations to get best fit
        #     temp = self.data_packets.data.set_index('specind')
        #     temp = temp.loc[mask]
        #     temp.reset_index(inplace=True)
        #
        #     W = np.zeros((len(data), len(self.modelfiles),
        #                   self.data_packets.data.Index0.max()+1))
        #
        #     spec_ind = [data.index.get_loc(i) for i in temp.specind.values]
        #     pack_ind = temp.index.to_list()
        #
        #     outputfiles = {ofile: i for i, ofile
        #                    in enumerate(self.modelfiles.keys())}
        #     oo = [outputfiles[ofile] for ofile in temp.outputfile]
        #     W[spec_ind, oo, temp.Index0.to_list()] = temp.weight
        #
        #     W2 = W.squeeze(axis=1)/W.mean()
        #     # assert np.all(W == W2.reshape(W.shape))
        #     assert W.shape[1] == 1
        #     scale_factor_, r, rank, s = lin.lstsq(W2, data.loc[:, 'radiance'].values)
        #     scale_factor_ /= scale_factor_[scale_factor_ > 0].mean()
        #     self.data_packets.data['scale_factor'] = (
        #         scale_factor_[self.data_packets.data.Index0.to_list()])
        #     self.data_packets.data['fit_weight'] = (self.data_packets.data.weight *
        #                                             self.data_packets.data.scale_factor)
        #
        #     surface_source['weight'] = scale_factor_[surface_source.Index.to_list()]
        #
        #     scaled = pd.read_pickle('surface_source_scaled.pkl')
        #
        #     from IPython import embed; embed()
        #     import sys; sys.exit()
        #
        #
        #     self.data_packets.data['scale_factor'] = (
        #         scale_factor_[self.data_packets.data['Index0'].to_list()])
        #
        #     # Determine surface weighting
        #
        #     surface_source.dropna(inplace=True)
        #     # surface_source = surface_source[surface_source.weight > 0]
        else:
            raise InputError('LOSResult.determine_source_from_data',
                             'Not a valid weighting method')

        initial = set(zip(self.data_packets.data.outputfile.to_list(),
                          self.data_packets.data.Index0.to_list(),
                          self.data_packets.data.frac0.to_list(),
                          self.data_packets.data.scale_factor.to_list()))
        frac0 = sum(x[2] * x[3] for x in initial)
        self.totalsource = frac0 * nsteps
        mod_rate = self.totalsource / endtime.value
        atoms_per_packet = 1e23 / mod_rate

        spec_group = self.data_packets.data.groupby('specind')
        fitrad_ = spec_group['fit_weight'].sum() * atoms_per_packet

        self.radiance.loc[fitrad_.index] = fitrad_
        self.radiance *= u.R

        # Create a new sourcemap
        source, xx, yy = np.histogram2d(surface_source['longitude'],
                                        surface_source['latitude'],
                                        weights=surface_source['weight'],
                                        range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
                                        bins=(NLONBINS, NLATBINS))
        source = source/np.cos(yy+(yy[1]-yy[0])/2.)[np.newaxis,:-1]
        source[:,[0,-1]] = 0
        
        v_source, v = np.histogram(surface_source['velocity'], bins=NVELBINS,
                                   range=[0, surface_source['velocity'].max()],
                                   weights=surface_source['weight'])
        v_source /= np.max(v_source)

        # packets available
        packets, _, _ = np.histogram2d(surface_source['longitude'],
                                       surface_source['latitude'],
                                       range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
                                       bins=(NLONBINS, NLATBINS))
        packets = packets/np.cos(yy+(yy[1]-yy[0])/2.)[np.newaxis,:-1]
        packets[:,[0,-1]] = 0
        v_packets, _ = np.histogram(surface_source['velocity'], bins=NVELBINS,
                                    range=[0, surface_source['velocity'].max()])
        v_packets = v_packets / np.max(v_packets)
        
        sourcemap = {'longitude': xx*u.rad,
                     'latitude': yy*u.rad,
                     'abundance': source,
                     'p_available': packets,
                     'velocity': v*u.km/u.s,
                     'vdist': v_source,
                     'v_available': v_packets,
                     'coordinate_system': 'solar-fixed'}
        self.sourcemap = sourcemap
        
    def simulate_data_from_inputs(self, inputs_, npackets, overwrite=False,
                                  packs_per_it=None, save_packets=False):
        """Given a set of inputs, determine what the spacecraft should see.
        
        Parameters
        ==========
        inputs_
            A nexoclom Input object or the name of an inputs file
        """
        if isinstance(inputs_, str):
            self.inputs = Input(inputs_)
        elif isinstance(inputs_, Input):
            self.inputs = copy.deepcopy(inputs_)
        else:
            raise InputError('nexoclom.LOSResult', 'Problem with the inputs.')
    
        # TAA needs to match the data
        self.inputs.geometry.taa = self.scdata.taa
        self.unit = u.def_unit('R_' + self.inputs.geometry.planet.object,
                               self.inputs.geometry.planet.radius)
    
        # If using a planet-fixed source map, need to set subsolarlon
        if ((self.inputs.spatialdist.type == 'surface map') and
            (self.inputs.spatialdist.coordinate_system == 'planet-fixed')):
            self.inputs.spatialdist.subsolarlon = self.scdata.subslong.median() * u.rad
        else:
            pass
    
        # Run the model
        self.fitted = False
        self.inputs.run(npackets, packs_per_it=packs_per_it, overwrite=overwrite)
        self.search_for_outputs()
        search_results = self.search()
    
        dist_from_plan = (self._data_setup()
                          if None in search_results.values()
                          else None)

        data = self.scdata.data
        for outputfile, search_result in search_results.items():
            data_packets_it = {'datafiles':[],
                               'totalsource':0.,
                               'radiance': pd.Series(index=self.scdata.data.index)}
            if search_result is None:
                # simulate the data
                output = Output.restore(outputfile)
                data_packets_it['totalsource'] = output.totalsource
                data_packets_it['out_idnum'] = output.idnum
                data_packets_it['outputfile'] = outputfile
                data_packets_it['endtime'] = output.inputs.options.endtime
                data_packets_it['nsteps'] = output.nsteps

                packets = copy.deepcopy(output.X)
                packets['frac0'] = output.X0.loc[packets['Index'], 'frac'].values
                packets['radvel_sun'] = (packets['vy'] +
                                         output.vrplanet.to(self.unit / u.s).value)
                self.oedge = output.inputs.options.outeredge * 2
        
                # Will base shadow on line of sight, not the packets
                out_of_shadow = np.ones(len(packets))
                self.packet_weighting(packets, out_of_shadow, output.aplanet)
        
                # This sets limits on regions where packets might be
                tree = self._tree(packets[xcols].values)
        
                vel_ = np.sqrt(output.X0.vx**2 + output.X0.vy**2 +
                               output.X0.vz**2) * output.inputs.geometry.planet.radius
                print(f'{data.shape[0]} spectra taken.')
                for i, spectrum in data.iterrows():
                    x_sc = spectrum[xcols].values.astype(float)
                    bore = spectrum[borecols].values.astype(float)
    
                    d_far = 30  # Furthest distance we need to look
                    x_far = x_sc + bore * d_far
                    while np.linalg.norm(x_far) > self.oedge:
                        d_far -= 0.1
                        x_far = x_sc + bore * d_far
    
                    t = [0.05]
                    while t[-1] < d_far:
                        t.append(t[-1] + t[-1] * np.sin(self.dphi))
                    t = np.array(t)
                    Xbore = x_sc[np.newaxis, :] + bore[np.newaxis, :] * t[:, np.newaxis]
    
                    wid = t * np.sin(self.dphi)
                    ind = np.concatenate(tree.query_radius(Xbore, wid))
                    ilocs = np.unique(ind).astype(int)
                    indicies = packets.iloc[ilocs].index
                    subset = packets.loc[indicies]
    
                    xpr = subset[xcols] - x_sc[np.newaxis, :]
                    rpr = np.sqrt(xpr['x'] * xpr['x'] +
                                  xpr['y'] * xpr['y'] +
                                  xpr['z'] * xpr['z'])
    
                    losrad = np.sum(xpr * bore[np.newaxis, :], axis=1)
                    inview = rpr < dist_from_plan[i]
    
                    if np.any(inview):
                        used_packets = subset[inview].index.to_list()
                        used_packets0 = subset.loc[inview, 'Index']
        
                        Apix = np.pi * (rpr[inview] * np.sin(self.dphi))**2 * (
                            self.unit.to(u.cm))**2
                        wtemp = subset.loc[inview, 'weight'] / Apix
                        if self.quantity == 'radiance':
                            # Determine if any packets are in shadow
                            # Projection of packet onto LOS
                            # Point along LOS the packet represents
                            losrad_ = losrad[inview].values
                            hit = (x_sc[np.newaxis, :] +
                                   bore[np.newaxis, :] * losrad_[:, np.newaxis])
                            rhohit = np.linalg.norm(hit[:, [0, 2]], axis=1)
                            out_of_shadow = (rhohit > 1) | (hit[:, 1] < 0)
                            wtemp *= out_of_shadow
            
                            rad = wtemp.sum()
                            data_packets_it['radiance'].loc[i] = rad
                            if save_packets:
                                # Save the weight information
                                rat = spectrum.radiance / rad if rad > 0 else 0.
                                processed = pd.DataFrame(
                                    {'outputfile':[outputfile for _ in used_packets],
                                     'specind':[i for _ in used_packets],
                                     'oind':used_packets,
                                     'weight':wtemp.values,
                                     'frac0':packets.loc[used_packets, 'frac0'].values,
                                     'Index0':used_packets0.values,
                                     'ratio':[rat for _ in used_packets]})
                                datafilename = f'temp/{os.path.basename(outputfile)}_{i}.parquet'
                                processed.to_parquet(datafilename)
                                del processed
                            else:
                                datafilename = None
                        else:
                            assert False, 'Other quantities not set up.'
                    else:
                        datafilename = None
    
                    if datafilename is not None:
                        data_packets_it['datafiles'].append(datafilename)
                    else:
                        pass
                    
                    ind = data.index.get_loc(i)
                    if (ind % (len(data)//10)) == 0:
                        print(f'Completed {ind+1} spectra')
                    else:
                        pass

                data_packets_it['datafiles'] = [file
                                                for file in data_packets_it['datafiles']
                                                if file is not None]
                
                if save_packets:
                    data_packets_df_it = dd.read_parquet(data_packets_it['datafiles'])
                    ind0 = data_packets_df_it.Index0.unique().compute().values
                    surface_it = pd.DataFrame(
                        {'outputfile': [outputfile for _ in ind0],
                         'Index0': ind0,
                         'longitude':output.X0.loc[ind0, 'longitude'].values,
                         'latitude':output.X0.loc[ind0, 'latitude'].values,
                         'velocity':vel_.loc[ind0]})
                else:
                    surface_it = None
                
                savefile, savefile_surf = self.save(data_packets_it, surface_it)
                self.modelfiles[outputfile] = savefile
                self.surfacefiles[outputfile] = savefile_surf
                del output
                del packets
                del surface_it
            else:
                data_packets_it = self.restore(search_result[1])
                self.modelfiles[outputfile] = search_result[1]
                self.surfacefiles[outputfile] = search_result[2]
                self.radiance += data_packets_it['radiance']

            self.radiance += data_packets_it['radiance']
            self.data_packets['datafiles'].extend(data_packets_it['datafiles'])
            self.data_packets['totalsource'] += data_packets_it['totalsource']
            del data_packets_it

        mod_rate = self.data_packets['totalsource'] / self.inputs.options.endtime.value
        atoms_per_packet = 1e23 / mod_rate
        self.radiance *= atoms_per_packet
        
        # data_packets_df = dd.read_parquet(self.data_packets['datafiles'])
        # rad_ = data_packets_df.groupby('specind').weight.sum().compute() * atoms_per_packet
        # rad_ = spec_group['weight'].sum() * atoms_per_packet
        # self.radiance.loc[rad_.index] = rad_
        self.radiance *= u.R

    def recompute_radiance(self, modnum, new_source):
        data = self.scdata.data
        radiance = pd.Series(index=data.index)

        modkey = f'model{modnum:02d}'
        maskkey = f'mask{modnum:02d}'
        
        self.modelfiles = self.scdata.model_info[modkey]['modelfiles']
        assert len(self.modelfiles) > 0
        self.outputfiles = self.modelfiles.keys()

        surface_source = pd.DataFrame(columns=['outputfile', 'Index', 'longitude',
                                               'latitude'])
        nsteps, endtime = 0, 0
        for outputfile, modefile in self.modelfiles.items():
            # Restore the unfit output file
            output = Output.restore(outputfile)
            self.totalsource += output.totalsource
            endtime = output.inputs.options.endtime
            nsteps = output.nsteps
            data_packets_it = self.restore(modefile)
            
            self.data_packets.data = self.data_packets.data.append(
                data_packets_it.data)
            self.data_packets.totalsource += data_packets_it.totalsource
            
            ind0 = data_packets_it.data.Index0.unique()
            surface_source = surface_source.append(pd.DataFrame(
                {'outputfile': [outputfile for _ in range(len(ind0))],
                 'Index': ind0,
                 'longitude': output.X0.loc[ind0, 'longitude'].values,
                 'latitude': output.X0.loc[ind0, 'latitude'].values}), ignore_index=True)

            del output
            del data_packets_it

        # Determine new weighting from given map
        new_weight_fn = RectBivariateSpline(new_source['longitude'][:-1].value,
                                            np.sin(new_source['latitude'][:-1].value),
                                            new_source['abundance'])
        scale_factor_ = new_weight_fn(surface_source.longitude,
                                      np.sin(surface_source.latitude),
                                      grid=False)
        scale_factor_ /= scale_factor_[scale_factor_ > 0].mean()
        surface_source['scale_factor'] = scale_factor_
        
        surface_ind = surface_source.set_index(['outputfile', 'Index'])
        ind0_ind = self.data_packets.data.set_index(['outputfile', 'Index0'])
        ind0_ind.loc[surface_ind.index, 'scale_factor'] = surface_ind.scale_factor
        # ind0_ind = ind0_ind.dropna()
        
        self.data_packets.data = ind0_ind.reset_index()
        self.data_packets.data['new_weight'] = (self.data_packets.data.weight *
                                                self.data_packets.data.scale_factor)
        
        initial = set(zip(self.data_packets.data.outputfile.to_list(),
                          self.data_packets.data.Index0.to_list(),
                          self.data_packets.data.frac0.to_list(),
                          self.data_packets.data.scale_factor.to_list()))
        frac0 = sum(x[2] * x[3] for x in initial)
        self.totalsource = frac0 * nsteps
        mod_rate = self.totalsource / endtime.value
        atoms_per_packet = 1e23 / mod_rate

        spec_group = self.data_packets.data.groupby('specind')
        fitrad_ = spec_group['new_weight'].sum() * atoms_per_packet
        # fitrad = spec_group['weight'].sum() * atoms_per_packet

        self.radiance.loc[fitrad_.index] = fitrad_
        self.radiance *= u.R


# if __name__ == '__main__':
#     import matplotlib.pyplot as plt
#     hist, xx, yy = np.histogram2d(surface_ind.longitude, surface_ind.latitude,
#                                   weights=surface_ind.scale_factor,
#                                   range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
#                                   bins=(NLONBINS, NLATBINS))
#     hist = hist / np.cos(yy + (yy[1] - yy[0]) / 2.)[np.newaxis, :-1]
#     hist *= new_source['abundance'].mean()/hist.mean()
#
#     fig, ax = plt.subplots(1, 2)
#     ax[0].imshow(new_source['abundance'], vmin=0, vmax=100)
#     ax[1].imshow(hist, vmin=0, vmax=100)
#     plt.show()
#     #
#
#     plt.plot(np.sum(new_source['abundance'], axis=0))
#     plt.plot(np.sum(hist, axis=0))
#     plt.show()
#     plt.plot(np.sum(new_source['abundance'], axis=1))
#     plt.plot(np.sum(hist, axis=1))
#     plt.show()
#     #
#     # from IPython import embed; embed()
#     # import sys; sys.exit()



import numpy as np


def smooth(array, num=7, method='mean', wrap=False):
    """
    :param array: Numpy array of numbers
    :param num: Smoothing full-width (should be odd). The smoothing window is
        [i-num/2, i + num/2].
    :param method: 'mean' or 'median'. Default = 'mean'
    :param wrap: If True, the smoothing window will treat the function as cyclical.
        If False (default), the smoothing window cuts off at the edges. For the
        first point, the smoothing window is [0, num/2]
    :return: Smoothed array
    """
    methods = {'mean': np.mean, 'median': np.median}
    func = methods.get(method, None)
    if isinstance(array, np.ndarray) and (func is not None):
        new_array = np.zeros_like(array)
        wid = int(num)//2
        for i in range(array.shape[0]):
            inds = np.linspace(i-wid, i+wid, wid*2+1, dtype=int)
            inds[inds >= len(array)] = inds[inds >= len(array)] - len(array)
            if not wrap:
                inds = inds[(inds >= 0) & (inds < len(array))]
            new_array[i] = func(array[inds])
            
        return new_array
    else:
        raise TypeError

# def smooth_sphere(array, longitude, latitude, dphi=np.deg2rad(5), method='mean'):
#     dlon, dlat = (longitude[1] - longitude[0]) / 2, (latitude[1] - latitude[0]) / 2
#     longitude_, latitude_ = np.meshgrid(longitude[:-1] + dlon,
#                                         latitude[:-1] + dlat)
#
#     ptsx = (np.cos(longitude_) * np.cos(latitude_)).flatten()
#     ptsy = (np.sin(longitude_) * np.cos(latitude_)).flatten()
#     ptsz = np.sin(latitude_).flatten()
#     array_ = array.flatten()
#     result = np.zeros_like(array_)
#
#     methods = {'mean': np.mean, 'median': np.median}
#     func = methods.get(method, None)
#     if isinstance(array, np.ndarray) and (func is not None):
#         for i, X in enumerate(zip(ptsx, ptsy, ptsz)):
#             # Compute angle between (x, y, z) and each point in the grid
#             AdotB = ptsx * X[0] + ptsy * X[1] + ptsz * X[2]
#             AdotB[AdotB > 1] = 1
#             AdotB[AdotB < -1] = -1
#             phi = np.arccos(AdotB)
#             use = phi < dphi
#             result[i] = func(array_[use])
#     else:
#         raise TypeError
#
#     return result.reshape(array.shape)





import os
import os.path
import pandas as pd
import numpy as np
import pickle
import random
import astropy.units as u
import sqlalchemy as sqla
import sqlalchemy.dialects.postgresql as pg

import nexoclom.math as mathMB
from nexoclom.solarsystem import planet_dist
from nexoclom.atomicdata import RadPresConst
from nexoclom.modelcode.satellite_initial_positions import satellite_initial_positions
from nexoclom.modelcode.LossInfo import LossInfo
from nexoclom.modelcode.rk5 import rk5
from nexoclom.modelcode.bouncepackets import bouncepackets
from nexoclom.modelcode.source_distribution import (surface_distribution,
                                                    speed_distribution,
                                                    angular_distribution)
from nexoclom.modelcode.SurfaceInteraction import SurfaceInteraction


class Output:
    def __init__(self, inputs, npackets, compress=True, logger=None):
        """Determine and store packet trajectories.
        
        **Parameters**
        
        inputs
            An Input object with the run parameters.
            
        npackets
            Number of packets to run.
        
        compress
            Remove packets with frac=0 from the outputs to reduce file size.
            Default = True
            
        **Class Attributes**
        
        x0, y0, z0
        
        f0
        
        vx0, vy0, vz0
        
        phi0, lat0, lon0
        
        time, x, y, z, vx, vy, vz
        Index, npackets, totalsource
        
        inputs
            The inputs used for the simulation
            
        logfile
            Path to file with output log
            
        compress
            Whether output is compressed.
        
        unit
            Basic length unit used. Equal to radius of central planet.
        
        GM
            GM_planet in units of R_planet/s**2
            
        aplanet
            Distance of planet from the Sun in AU
         
        vrplanet
            Radial velocity of planet relative to the Sun in R_planet/s
        
        radpres
            Radiation pressure object containing acceleration as funtion
            of velocity in units of R_planet/s**2 and R_planet/s
        """
        # if logger is None:
        #     logger = logging.getLogger()
        #     logger.setLevel(logging.INFO)
        #     out_handler = logging.StreamHandler(sys.stdout)
        #     logger.addHandler(out_handler)
        #     fmt = logging.Formatter('%(levelname)s: %(msg)s')
        #     out_handler.setFormatter(fmt)
        # else:
        #     pass
        # self.logger = logger

        self.inputs = inputs
        self.planet = inputs.geometry.planet
        
        # initialize the random generator
        self.randgen = np.random.default_rng()
        
        # Not implemented yet.
        assert self.inputs.geometry.type != 'geometry with time', (
            'Initialization with time stamp not implemented yet.')

        # Keep track of whether output is compressed
        self.compress = compress

        # Determine spatial unit
        self.unit = u.def_unit('R_' + self.planet.object, self.planet.radius)

        # Change unit for GM
        self.GM = self.planet.GM.to(self.unit**3/u.s**2).value

        # Determine distance and radial velocity of planet relative to the Sun
        r, v_r = planet_dist(self.planet, self.inputs.geometry.taa)
        self.aplanet = r.value
        self.vrplanet = v_r.to(self.unit/u.s).value

        # Find the default reactions and datasets
        if inputs.options.lifetime.value <= 0:
            self.loss_info = LossInfo(inputs.options.species,
                                      inputs.options.lifetime,
                                      self.aplanet)
        else:
            self.loss_info = None

        # Set up the radiation pressure
        if inputs.forces.radpres:
            radpres = RadPresConst(inputs.options.species,
                                   self.aplanet)
            radpres.velocity = radpres.velocity.to(self.unit/u.s).value
            radpres.accel = radpres.accel.to(self.unit/u.s**2).value
            self.radpres = radpres
        else:
            self.radpres = None

        # set up surface accommodation + maybe other things if needed
        if (('stickcoef' not in inputs.surfaceinteraction.__dict__) or
            (inputs.surfaceinteraction.stickcoef != 1)):
            self.surfaceint = SurfaceInteraction(inputs,
                                                 nt=201, nv=101, nprob=101)

        # Define the time that packets will run
        if inputs.options.step_size > 0:
            time = np.ones(npackets) * inputs.options.endtime
        else:
            time = self.randgen.random(npackets) * inputs.options.endtime

        self.X0 = pd.DataFrame()
        self.X0['time'] = time.value

        # Define the fractional content
        self.X0['frac'] = np.ones(npackets)

        self.npackets = npackets
        self.totalsource = self.X0['frac'].sum()

        # Determine initial satellite positions if necessary
        if self.planet.moons is not None:
            assert False, 'Not set up'
            sat_init_pos = satellite_initial_positions(inputs)
        else:
            pass

        # Determine starting location for each packet
        if self.inputs.spatialdist.type in ('uniform', 'surface map',
                                            'surface spot'):
            surface_distribution(self)
        else:
            assert 0, 'Not a valid spatial distribution type'
        
        # Determine inital speed for each packet
        speed_distribution(self)
        
        # Choose direction for each packet
        angular_distribution(self)

        # Rotate everything to proper position for running the model
        if (self.inputs.geometry.planet.object !=
            self.inputs.geometry.startpoint):
            assert 0, 'Not set up yet'
        else:
            pass
        
        # Reorder the dataframe columns
        cols = ['time', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'frac', 'v',
                'longitude', 'latitude', 'local_time']
        self.X0 = self.X0[cols]
        
        # Integrate the packets forward
        if self.inputs.options.step_size == 0:
            print('Running variable step size integrator.')
            self.X = self.X0.drop(['longitude', 'latitude', 'localtime'], axis=1)
            self.X['lossfrac'] = np.zeros(npackets)
            self.variable_step_size_driver()
        else:
            print('Running constant step size integrator.')
            self.constant_step_size_driver()
            
        self.save()

    def __str__(self):
        print('Contents of output:')
        print('\tPlanet = {}'.format(self.planet.object))
        print('\ta_planet = {}'.format(self.aplanet))
        print('\tvr_planet = {}'.format(self.vrplanet))
        print('\tNumber of Packets: {}'.format(self.npackets))
#        print('\tUnits of time: {}'.format(self.time.unit))
#        print('\tUnits of distance: {}'.format(self.X0.unit))
#        print('\tUnits of velocity: {}'.format(self.V0.unit))
        return ''

    def __len__(self):
        return self.npackets

    def __getitem__(self, keys):
        self.X = self.X.iloc[keys]

    def variable_step_size_driver(self):
        # Set up the step sizes
        count = 0  # Number of steps taken

        # These control how quickly the stepsize is increased or decreased
        # between iterations
        safety = 0.95
        shrink = -0.25
        grow = -0.2

        # yscale = scaling parameter for each variable
        #     x,y,z ~ R_plan
        #     vx, vy, vz ~ 1 km/s (1/R_plan R_plan/s)
        #     frac ~ exp(-t/lifetime) ~ mean(frac)
        rest = self.inputs.options.resolution
        resx = self.inputs.options.resolution
        resv = 0.1*self.inputs.options.resolution
        resf = self.inputs.options.resolution

        #########################################################
        # Keep taking RK steps until every packet has reached the
        # time of "image taken"
        #########################################################

        # initial step size
        step_size = np.zeros(self.npackets) + 1000.
        cols = ['time', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'frac']
        moretogo = (self.X['time'] > rest) & (self.X['frac'] > 0.)
        while moretogo.any():
            # Save old values
            # This is used for determining if anything hits the rings
            Xtodo = self.X[cols][moretogo].values
            step = step_size[moretogo]
            Xold = Xtodo.copy()
            
            if np.any(step < 0):
                # self.logger.error('Negative values of h '
                #                   'in variable_step_size_dirver')
                print('Negative values of h '
                      'in variable_step_size_dirver')
                assert 0, '\n\tNegative values of step_size'
            else:
                pass

            # Adjust stepsize to be no more than time remaining
            step = np.minimum(Xtodo[:,0], step)

            # Run the rk5 step
            Xnext, delta = rk5(self, Xtodo, step)

            # Do the error check
            # scale = a_tol + |y|*r_tol
            #   for x: a_tol = r_tol = resolution
            #   for v: a_tol = r_tol = resolution/10.-require v more precise
            #   for f: a_tol = 0.01, r_tol = 0 -> frac tol = 1%
            scalex = resx + np.abs(Xnext[:,1:4])*resx
            scalev = resv + np.abs(Xnext[:,4:7])*resv
            scalef = resf + np.abs(Xnext[:,7])*resf

            # Difference relative to acceptable difference
            delta[:,1:4] /= scalex
            delta[:,4:7] /= scalev
            delta[:,7] /= scalef

            # Maximum error for each packet
            errmax = delta.max(axis=1)

            # error check
            assert np.all(np.isfinite(errmax)), '\n\tInfinite values of emax'

            # Make sure no negative frac
            assert not np.any((Xnext[:,7] < 0) & (errmax < 1)),(
                'Found new values of frac that are negative')

            # Make sure frac doesn't increase
            errmax[(Xnext[:,7]-Xtodo[:,7] > scalef) & (errmax > 1)] = 1.1

            # Check where difference is very small. Adjust step size
            noerr = errmax < 1e-7
            errmax[noerr] = 1
            step[noerr] *= 10

            # Put the post-step values in
            g = errmax < 1.0
            b = errmax >= 1.0

            if np.any(g):
                Ximpcheck = Xnext[g,:]
                step_ = safety*step[g]*errmax[g]**grow

                # Impact Check
                tempR = np.linalg.norm(Ximpcheck[:,1:4], axis=1)
                hitplanet = (tempR - 1.) < 0
                if np.any(hitplanet):
                    if ((self.inputs.surfaceinteraction.sticktype == 'constant')
                        and (self.inputs.surfaceinteraction.stickcoef == 1.)):
                        Xnext[hitplanet, 7] = 0.
                    else:
                        bouncepackets(self, Xnext[hitplanet, :],
                                      tempR[hitplanet])
                else:
                    pass

                # Check for escape
                Ximpcheck[tempR > self.inputs.options.outeredge,7] = 0

                # Check for vanishing
                Ximpcheck[Ximpcheck[:,7] < 1e-10, 7] = 0.

                # set remaining time = 0 for packets that are done
                Ximpcheck[Ximpcheck[:,7] == 0, 0] = 0.

                # Put new values into arrays
                #Xnext[g,:] = Ximpcheck
                Xtodo[g] = Ximpcheck
                step[g] = step_
            else: pass

            if np.any(b):
                # Don't adjust the bad value, but do fix the stepsize
                step_ = safety*step[b]*errmax[b]**shrink
                assert np.all(np.isfinite(step_)), (
                    '\n\tInfinite values of step_size')

                # Don't let step size drop below 1/10th previous step size
                step[b] = np.maximum(step_, 0.1*step[b])

            assert np.all(step >= 0), '\n\tNegative values of step_size'

            # Insert back into the original arrays
            self.X.loc[moretogo,cols] = Xtodo
            self.X.loc[moretogo,'lossfrac'] += Xold[:,7] - Xtodo[:,7]
            step_size[moretogo] = step

            # Find which packets still need to run
            moretogo = (self.X['time'] > rest) & (self.X['frac'] > 0.)
            if count % 100 == 0:
                print(f'Step {count}. {np.sum(moretogo)} more to go\n'
                      f'\tstep_size: {mathMB.minmaxmean(step_size)}')
            count += 1

        # Add units back in
        self.aplanet *= u.au
        self.vrplanet *= self.unit/u.s
        self.vrplanet = self.vrplanet.to(u.km/u.s)
        self.GM *= self.unit**3/u.s**2

    def constant_step_size_driver(self):
        # Arrays to store the outputs
        cols = ['time', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'frac']

        #  step size and counters
        step_size = np.zeros(self.npackets) + self.inputs.options.step_size
        
        self.nsteps = int(np.ceil(self.inputs.options.endtime.value/step_size[0]
                             + 1))
        results = np.zeros((self.npackets,8,self.nsteps))
        results[:,:,0] = self.X0[cols]
        lossfrac = np.ndarray((self.npackets,self.nsteps))

        curtime = self.inputs.options.endtime.value
        ct = 1
        moretogo = results[:,7,0] > 0
        while (curtime > 0) and (moretogo.any()):
            Xtodo = results[moretogo,:,ct-1]
            step = step_size[moretogo]

            assert np.all(Xtodo[:,7] > 0)
            assert np.all(np.isfinite(Xtodo))

            # Run the rk5 step
            Xnext, _ = rk5(self, Xtodo, step)

            # Check for surface impacts
            tempR = np.linalg.norm(Xnext[:,1:4], axis=1)
            hitplanet = (tempR - 1.) < 0

            if np.any(hitplanet):
                if ((self.inputs.surfaceinteraction.sticktype == 'constant')
                    and (self.inputs.surfaceinteraction.stickcoef == 1.)):
                        Xnext[hitplanet,7] = 0.
                else:
                    bouncepackets(self, Xnext, tempR, hitplanet)
            else:
                pass

            # Check for escape
            Xnext[tempR > self.inputs.options.outeredge,7] = 0
            
            # Check for vanishing
            Xnext[Xnext[:, 7] < 1e-10, 7] = 0.

            # set remaining time = 0 for packets that are done
            Xnext[Xnext[:, 7] == 0, 0] = 0.

            # Put new values back into the original array
            results[moretogo,:,ct] = Xnext
            lossfrac[moretogo,ct] = (lossfrac[moretogo,ct-1] +
                results[moretogo,7,ct-1] - results[moretogo,7,ct])
            
            # Check to see what still needs to be done
            moretogo = results[:,7,ct] > 0

            if (ct % 100) == 0:
                print(ct, curtime, int(np.sum(moretogo)))

            # Update the times
            ct += 1
            curtime -= step_size[0]

        # Put everything back into output
        self.totalsource *= self.nsteps
        X = pd.DataFrame()
        index = np.mgrid[:self.npackets, :self.nsteps]
        npackets = self.npackets * self.nsteps
        X['Index'] = index[0,:,:].reshape(npackets)
        X['time'] = results[:,0,:].reshape(npackets)
        X['x'] = results[:,1,:].reshape(npackets)
        X['y'] = results[:,2,:].reshape(npackets)
        X['z'] = results[:,3,:].reshape(npackets)
        X['vx'] = results[:,4,:].reshape(npackets)
        X['vy'] = results[:,5,:].reshape(npackets)
        X['vz'] = results[:,6,:].reshape(npackets)
        X['frac'] = results[:,7,:].reshape(npackets)
        X['lossfrac'] = lossfrac.reshape(npackets)
        self.X = X

        # Add units back in
        self.aplanet *= u.au
        self.vrplanet *= self.unit/u.s
        self.vrplanet = self.vrplanet.to(u.km/u.s)
        self.GM *= self.unit**3/u.s**2

    def make_filename(self):
        """Determine filename for output."""
        # TAA for observation
        if self.planet.object == 'Mercury':
            taastr = '{:03.0f}'.format(
                      np.round(self.inputs.geometry.taa.to(u.deg).value))
        else:
            assert 0, 'Filename not set up for anything but Mercury'

        # Come up with a path name
        pathname = os.path.join(self.inputs.config.savepath,
                                self.planet.object,
                                self.inputs.options.species,
                                self.inputs.spatialdist.type,
                                self.inputs.speeddist.type,
                                taastr)

        # Make the path if necessary
        if os.path.exists(pathname) is False:
            os.makedirs(pathname)
        numstr = '{:010d}'.format(self.idnum)
        self.filename = os.path.join(pathname, f'{numstr}.pkl')

    def save(self):
        """Add output to database and save as a pickle."""
        geo_id = self.inputs.geometry.insert()
        sint_id = self.inputs.surfaceinteraction.insert()
        for_id = self.inputs.forces.insert()
        spat_id = self.inputs.spatialdist.insert()
        spd_id = self.inputs.speeddist.insert()
        ang_id = self.inputs.angulardist.insert()
        opt_id = self.inputs.options.insert()
        
        engine = self.inputs.config.create_engine()
        metadata_obj = sqla.MetaData()
        table = sqla.Table("outputfile", metadata_obj, autoload_with=engine)
        
        insert_stmt = pg.insert(table).values(
            filename = None,
            npackets = self.npackets,
            totalsource = self.totalsource,
            geo_type = self.inputs.geometry.type,
            geo_id = geo_id,
            sint_type = self.inputs.surfaceinteraction.sticktype,
            sint_id = sint_id,
            force_id = for_id,
            spatdist_type = self.inputs.spatialdist.type,
            spatdist_id = spat_id,
            spddist_type = self.inputs.speeddist.type,
            spddist_id = spd_id,
            angdist_type = self.inputs.angulardist.type,
            angdist_id = ang_id,
            opt_id = opt_id)
        
        with engine.connect() as con:
            result = con.execute(insert_stmt)
            con.commit()
            
        self.idnum = result.inserted_primary_key[0]
        self.make_filename()
        update = sqla.update(table).where(table.columns.idnum == self.idnum).values(
            filename=self.filename)
        with engine.connect() as con:
            con.execute(update)
            con.commit()
            
        # Remove frac = 0
        if self.compress:
            self.X = self.X[self.X.frac > 0]
        else:
            pass
        
        # Convert to 32 bit
        for column in self.X0:
            if self.X0[column].dtype == np.int64:
                self.X0[column] = self.X0[column].astype(np.int32)
            elif self.X0[column].dtype == np.float64:
                self.X0[column] = self.X0[column].astype(np.float32)
            else:
                pass

        for column in self.X:
            if self.X[column].dtype == np.int64:
                self.X[column] = self.X[column].astype(np.int32)
            elif self.X[column].dtype == np.float64:
                self.X[column] = self.X[column].astype(np.float32)
            else:
                pass

        # Save output as a pickle
        print(f'Saving file {self.filename}')
        if self.inputs.surfaceinteraction.sticktype == 'temperature dependent':
            self.surfaceint.stickcoef = 'FUNCTION'
            
        with open(self.filename, 'wb') as f:
            pickle.dump(self, f, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def restore(cls, filename):
        output = pickle.load(open(filename, 'rb'))
        return output



"""``g_values`` - Routines related to g-values and radiation pressure
The g-value is the the product of the solar flux at the dopler-shifted
emission wavelength and the scattering probability per atom. See
`Killen, R.M. et al., Icarus 209, 75–87, 2009.
<http://dx.doi.org/10.1016/j.icarus.2010.02.018.>`_ for details on calculating
g-values for important species in Mercury's atmosphere.

The radiation acceleration is given by
:math:`a_{rad} = h g/m \lambda`,
where h is Plank's constant, g is the g-value as a function of radial
velocity, m is the mass of the accelerated species, and λ is the wavelength
of the absorbed photon.
"""
import os
import numpy as np
import pandas as pd
import astropy.units as u
from astropy import constants as const
from nexoclom.atomicdata.atomicmass import atomicmass
from nexoclom.math import interpu
from nexoclom import __file__ as basefile
# pylint: disable=no-member

class gValue:
    r"""Class containing g-value vs. velocity for a specified atom and
    transition.

    **Parameters**

    sp
        atomic species

    wavelength
        Wavelength of the transition. Default=None.

    aplanet
        Distance from the Sun. Can be given as an astropy quantity with
        distance units or as a float assumed to be in AU. Default = 1 AU

    **Class Attributes**

    species
        The input species

    wavelength
        The input wavelength

    aplanet
        The input aplanet

    velocity
        Radial velocity deviation relative to the Sun in km/s.
        Positive values indicate
        motion away from the Sun. Given as a numpy array of astropy quantities

    g
        g-value as function of velocity in units 1/s.
    """
    def __init__(self, sp, wavelength, aplanet=1*u.au):
        self.species = sp

        try:
            self.wavelength = wavelength.to(u.AA)
        except:
            self.wavelength = wavelength * u.AA

        try:
            self.aplanet = aplanet.to(u.au)
        except:
            self.aplanet = aplanet * u.au

        gvalue_file = self.gvalue_filename()
        gvalues = pd.read_pickle(gvalue_file)

        gvalue = gvalues[(gvalues.species == sp) &
                         (gvalues.wavelength == wavelength)]

        if len(gvalue) == 0:
            self.velocity = np.array([0., 1.])*u.km/u.s
            self.g = np.array([0., 0.])/u.s
            self.filename = None
            self.reference = None
            print(f'Warning: g-values not found for species = {sp}')
        elif len(gvalue.filename.unique()) == 1:
            self.velocity = gvalue.velocity.values*u.km/u.s
            self.g = (gvalue.gvalue *
                      gvalue.refpoint**2/self.aplanet.value**2).values/u.s
            s = np.argsort(self.velocity)
            self.velocity, self.g = self.velocity[s], self.g[s]
            self.reference = gvalue.reference.unique()[0]
            self.filename = gvalue.filename.unique()[0]
        else:
            print('This should never happen')
            raise ValueError()

    @classmethod
    def gvalue_filename(self):
        return os.path.join(os.path.dirname(basefile), 'data', 'g-values', 
                            'g-values.pkl')


class RadPresConst:
    r"""Class containing radial acceleration vs. velocity for a specified atom.

    **Parameters**

    sp
        atomic species

    aplanet
        Distance from the Sun. Can be given as an astropy quantity with
        distance units or as a float assumed to be in AU. Default = 1 AU

    database
        Database containing solar system information. Default =
        `thesolarsystem` which probably shouldn't be overridden.

    **Class Attributes**

    species
        The input species

    aplanet
        The input distance from the Sun

    velocity
        Radial velocity deviation relative to the Sun in km/s.
        Positive values indicate
        motion away from the Sun. Given as a numpy array of astropy quantities

    accel
        Radial acceleration vs. velocity with units km/s**2.
    """
    def __init__(self, species, aplanet):
        self.species = species
        if isinstance(aplanet, type(1*u.au)):
            self.aplanet = aplanet
        else:
            self.aplanet = aplanet * u.au

        gvalues = pd.read_pickle(gValue.gvalue_filename())

        if species in gvalues.species.values:
            subset = gvalues.loc[gvalues.species == species]
            self.wavelength = np.array(sorted(subset.wavelength.unique())) * u.AA
            self.velocity = np.array(sorted(subset.velocity.unique())) * u.km/u.s

            # Interpolate gvalues to full velocity set and compute rad pres
            rpres = np.zeros_like(self.velocity)/u.s
            for wave in self.wavelength:
                gval = gValue(species, wave, aplanet)
                g_ = interpu(self.velocity, gval.velocity, gval.g)
                rpres_ = const.h/atomicmass(species)/wave * g_
                rpres += rpres_.to(u.km/u.s**2)
                
            self.accel = rpres
        else:
            self.velocity = np.array([0., 1.])*u.km/u.s
            self.accel = np.array([0., 0.])*u.km/u.s**2
            print(f'Warning: g-values not found for species = {species}')


import os
import numpy as np
import pickle
import astropy.units as u
import astropy.constants as const
import nexoclom.math as mathMB
from nexoclom.atomicdata import atomicmass
import nexoclom.math.distributions as distributions


def xyz_from_lonlat(lon, lat, isplan, exobase):
    if isplan:
        # Starting at a planet
        # 0 deg longitude = subsolar pt. = (0, -1, 0)
        # 90 deg longitude = dusk pt. = (1, 0, 0)
        # 270 deg longitude = dawn pt. = (-1, 0, 0)
        x0 = exobase * np.sin(lon) * np.cos(lat)
        y0 = -exobase * np.cos(lon) * np.cos(lat)
        z0 = exobase * np.sin(lat)
    else:
        # Starting at a satellite
        # 0 deg longitude = sub-planet pt. = (0, -1, 0)
        # 90 deg longitude = leading pt. = (-1, 0, 0)
        # 270 deg longitude = trailing pt. = (1, 0, 0)
        x0 = -exobase * np.sin(lon) * np.cos(lat)
        y0 = -exobase * np.cos(lon) * np.cos(lat)
        z0 = exobase * np.sin(lat)

    X0 = np.array([x0, y0, z0])

    # Error checking
    assert np.all(np.isfinite(X0)), 'Non-Finite values of X0'
    return X0


def surface_distribution(outputs):
    """ Distribute packets on a sphere with radius r = SpatialDist.exobase

    Returns (x0, y0, z0, lon0, lat0)
    for satellites, assumes satellite is at phi=0
    """

    spatialdist = outputs.inputs.spatialdist
    npack = outputs.npackets
    
    if spatialdist.type == 'uniform':
        # Choose the latitude: f(lat) = cos(lat)
        # lat0 = spatialdist.latitude
        # ll = (np.sin(lat0[0]), np.sin(lat0[1]))
        ll = tuple(map(np.sin, spatialdist.latitude))
        sinlat = ll[0] + (ll[1]-ll[0]) * outputs.randgen.random(npack)
        lat = np.arcsin(sinlat)
    
        # Choose the longitude: f(lon) = 1/(lonmax-lonmin)
        lon0 = spatialdist.longitude
        if lon0[0] > lon0[1]:
            lon0 = [lon0[0], lon0[1]+2*np.pi*u.rad]
        else:
            pass
        lon = ((lon0[0] + (lon0[1]-lon0[0]) * outputs.randgen.random(npack)) %
               (2*np.pi*u.rad))
    elif spatialdist.type == 'surface map':
        # Choose lon, lat based on predetermined map
        if spatialdist.mapfile == 'default':
            mapfile = os.path.join(os.path.dirname(__file__), 'data',
                f'{outputs.inputs.options.species}_surface_composition.pkl')
            with open(mapfile, 'rb') as mfile:
                sourcemap = pickle.load(mfile)
        elif spatialdist.mapfile.endswith('.pkl'):
            with open(spatialdist.mapfile, 'rb') as mfile:
                sourcemap = pickle.load(mfile)
        elif spatialdist.mapfile.endswith('.sav'):
            from scipy.io import readsav
            sourcemap_ = readsav(spatialdist.mapfile)['sourcemap']
            sourcemap = {'longitude':sourcemap_['longitude'][0]*u.rad,
                         'latitude':sourcemap_['latitude'][0]*u.rad,
                         'abundance':sourcemap_['map'][0].transpose(),
                         'coordinate_system':str(sourcemap_['coordinate_system'][0])}
        else:
            assert 0, 'Mapfile is the wrong format.'

        lon, lat = mathMB.random_deviates_2d(sourcemap['abundance'],
                                             sourcemap['longitude'],
                                             np.sin(sourcemap['latitude']),
                                             npack)
        lat = np.arcsin(lat)
        
        if (('planet' in sourcemap['coordinate_system']) and
            (outputs.inputs.spatialdist.subsolarlon is not None)):
            # Need to rotate to model coordinate system
            lon = ((outputs.inputs.spatialdist.subsolarlon.value - lon +
                    2*np.pi) % (2*np.pi))
        elif ('planet' in sourcemap['coordinate_system']):
            raise ValueError('inputs.spatialdist.subsolarlon is None')
        else:
            pass
    elif spatialdist.type == 'surface spot':
        lon0 = spatialdist.longitude
        lat0 = spatialdist.latitude
        sigma0 = spatialdist.sigma
    
        spot0 = ((np.sin(lon0)*np.cos(lat0)).value,
                 (-np.cos(lon0)*np.cos(lat0)).value,
                 (np.sin(lat0)).value)
        longitude = np.linspace(0, 2*np.pi, 361)*u.rad
        latitude = np.linspace(-np.pi/2, np.pi/2, 181)*u.rad
    
        ptsx = np.outer(np.sin(longitude.value), np.cos(latitude.value))
        ptsy = -np.outer(np.cos(longitude.value), np.cos(latitude.value))
        ptsz = -np.outer(np.ones_like(longitude.value), np.sin(latitude.value))
    
        cosphi = ptsx*spot0[0]+ptsy*spot0[1]+ptsz*spot0[2]
        cosphi[cosphi > 1] = 1
        cosphi[cosphi < -1] = -1
        phi = np.arccos(cosphi)
        sourcemap = np.exp(-phi/sigma0.value)
    
        lon, lat = mathMB.random_deviates_2d(sourcemap, longitude,
                                             np.sin(latitude), npack)
        lat = np.arcsin(lat)
    else:
        assert 0, "Can't get here"

    X_ = xyz_from_lonlat(lon, lat,
                         outputs.inputs.geometry.planet.type == 'Planet',
                         spatialdist.exobase)
    outputs.X0['x'] = X_[0,:]
    outputs.X0['y'] = X_[1,:]
    outputs.X0['z'] = X_[2,:]
    outputs.X0['longitude'] = lon.value
    outputs.X0['latitude'] = lat.value
    local_time = (lon.value * 12/np.pi + 12) % 24
    outputs.X0['local_time'] = local_time


def speed_distribution(outputs):
    speeddist = outputs.inputs.speeddist
    npackets = outputs.npackets

    if speeddist.type.lower() == 'gaussian':
        if speeddist.sigma == 0.:
            v0 = np.zeros(npackets)*u.km/u.s + speeddist.vprob
        else:
            v0 = (outputs.randgen.standard_normal(npackets) *
                  speeddist.sigma.value + speeddist.vprob.value)
            v0 *= speeddist.vprob.unit
    elif speeddist.type == 'sputtering':
        velocity = np.linspace(.1, 50, 5000)*u.km/u.s
        f_v = distributions.sputdist(velocity, speeddist.U, speeddist.alpha,
            speeddist.beta, outputs.inputs.options.species)
        v0 = (mathMB.random_deviates_1d(velocity.value, f_v, npackets) *
              velocity.unit)
    elif speeddist.type == 'maxwellian':
        if speeddist.temperature != 0*u.K:
            # Use a constant temperature
            amass = atomicmass(outputs.inputs.options.species)
            v_th = np.sqrt(2*speeddist.temperature*const.k_B/amass)
            v_th = v_th.to(u.km/u.s)
            velocity = np.linspace(0.1*u.km/u.s, v_th*5, 5000)
            f_v = distributions.MaxwellianDist(velocity, speeddist.temperature,
                                               outputs.inputs.options.species)
            v0 = (mathMB.random_deviates_1d(velocity.value, f_v, npackets) *
                  velocity.unit)
        else:
            # Use a surface temperature map
            # Need to write this
            assert 0, 'Not implemented yet'
    elif speeddist.type == 'flat':
        v0 = (outputs.randgen.random(npackets)*2*speeddist.delv +
              speeddist.vprob - speeddist.delv)
    elif speeddist.type == 'user defined':
        source = pickle.load(open(speeddist.vdistfile, 'rb'))
        v0 = mathMB.random_deviates_1d(source['velocity'].value, source['vdist'],
                                       npackets) * source['velocity'].unit
    else:
        # Need to add more distributions
        assert 0, 'Distribtuion does not exist'
        
    v0 = v0.to(outputs.unit/u.s)
    outputs.X0['v'] = v0.value

    assert np.all(np.isfinite(v0)), 'Infinite values for v0'

    return v0


def angular_distribution(outputs):
    npackets = outputs.npackets

    angulardist = outputs.inputs.angulardist

    if angulardist.type == 'none':
        return
    elif angulardist.type == 'radial':
        # All packets going radially outward
        alt = (np.zeros(npackets) + np.pi/2.) * u.rad
        az = np.zeros(npackets) * u.rad
    elif angulardist.type == 'isotropic':
        # Choose the altitude -- f(alt) = cos(alt)
        alt0 = angulardist.altitude
        aa = (np.sin(alt0[0]), np.sin(alt0[1]))
        sinalt = outputs.randgen.random(npackets) * (aa[1] - aa[0]) + aa[0]
        alt = np.arcsin(sinalt)

        # Choose the azimuth -- f(az) = 1/(azmax-azmin)
        az0, az1 = angulardist.azimuth
        m = (az0, az1) if az0 < az1 else (az1, az0+2*np.pi)
        az = m[0] + (m[1]-m[0])*outputs.randgen.random(npackets)
    else:
        assert 0, 'Angular Distribution not defined.'

    # Find the velocity components in coordinate system centered on packet
    v_rad = np.sin(alt.value)                 # Radial component of velocity
    v_tan0 = np.cos(alt.value) * np.cos(az.value)   # Component along latitude (points E)
    v_tan1 = np.cos(alt.value) * np.sin(az.value)   # Component along longitude (points N)

    # Now rotate to proper surface point
    # v_ren = M # v_xyz => v_xyz = invert(M) # v_ren
    X0 = outputs.X0.values
    x0, y0, z0 = X0[:,2], X0[:,3], X0[:,4]

    rad = np.array([x0, y0, z0]).transpose()
    east = np.array([y0, -x0, np.zeros_like(z0)]).transpose()
    north = np.array([-z0*x0, -z0*y0, x0**2+y0**2]).transpose()

    rad_ = np.linalg.norm(rad, axis=1)
    rad = rad/rad_[:,np.newaxis]
    east_ = np.linalg.norm(east, axis=1)
    east = east/east_[:,np.newaxis]
    north_ = np.linalg.norm(north, axis=1)
    north = north/north_[:,np.newaxis]

    v0 = (v_tan0[:,np.newaxis]*north + v_tan1[:,np.newaxis]*east +
          v_rad[:,np.newaxis]*rad)
    
    outputs.X0['vx'] = v0[:,0] * outputs.X0.v.values
    outputs.X0['vy'] = v0[:,1] * outputs.X0.v.values
    outputs.X0['vz'] = v0[:,2] * outputs.X0.v.values



import numpy as np
from nexoclom.modelcode.surface_temperature import surface_temperature


def rebound_direction(outputs, x):
    npackets = x.shape[0]

    # Choose the altitude -- f(alt) = cos(alt)
    sinalt = outputs.randgen.random(npackets)
    alt = np.arcsin(sinalt)

    # Choose the azimuth -- f(az) = 1/(azmax-azmin)
    az = 2*np.pi*outputs.randgen.random(npackets)

    # Find the velocity components in coordinate system centered on packet
    v_rad = np.sin(alt)                 # Radial component of velocity
    v_tan0 = np.cos(alt) * np.cos(az)   # Component along latitude (points E)
    v_tan1 = np.cos(alt) * np.sin(az)   # Component along longitude (points N)

    # Now rotate to proper surface point
    # v_ren = M # v_xyz => v_xyz = invert(M) # v_ren

    rad = x.copy()
    rad /= np.linalg.norm(rad, axis=1)[:,np.newaxis]
    east = np.array([x[:,1], -x[:,0], np.zeros(npackets)]).transpose()
    east /= np.linalg.norm(east, axis=1)[:,np.newaxis]
    north = np.array([-x[:,2]*x[:,0], -x[:,2]*x[:,1],
                      x[:,0]**2 + x[:,1]**2]).transpose()
    north /= np.linalg.norm(north, axis=1)[:,np.newaxis]
    
    direction = (v_tan0[:,np.newaxis]*north +
                 v_tan1[:,np.newaxis]*east +
                 v_rad[:,np.newaxis]*rad)
    assert np.all(np.isfinite(direction))
    
    return direction


def bouncepackets(outputs, Ximpcheck, r0, hitplanet):
    npackets = np.sum(hitplanet)
    # This will need to be rewritten for satellite impacts

    # Determine where packets hit surface
    Xtemp, rtemp = Ximpcheck[hitplanet,:], r0[hitplanet]
    a = np.sum(Xtemp[:,4:7]**2, axis=1)  # = vv02
    b = 2 * np.sum(Xtemp[:,1:4] * Xtemp[:,4:7], axis=1)
    c = np.sum(Xtemp[:,1:4]**2, axis=1) - 1.
    dd = b**2 - 4*a*c

    t0 = (-b - np.sqrt(b**2-4*a*c))/(2*a)
    t1 = (-b + np.sqrt(b**2-4*a*c))/(2*a)
    t = np.minimum(t0, t1) # t0 <= 0)*t0 + (t1 < 0)*t1

    # point where packet hit the surface
    Xtemp[:,1:4] = Xtemp[:,1:4] + Xtemp[:,4:7] * t[:,np.newaxis]
    assert np.all(np.isclose(np.linalg.norm(Xtemp[:,1:4], axis=1), 1.))

    # Determine impact velocity
    PE = 2*outputs.GM*(1./rtemp - 1)
    v_old2 = a + PE
    v_old2[v_old2 < 0] = 0.

    direction = rebound_direction(outputs, Xtemp[:,1:4])
    
    if outputs.inputs.surfaceinteraction.accomfactor == 0:
        v_new = np.sqrt(v_old2)
    else:
        lonhit = (np.arctan2(Xtemp[:,1], -Xtemp[:,2])+2*np.pi) % (2*np.pi)
        lathit = np.arcsin(Xtemp[:,3])
        
        surftemp = surface_temperature(outputs.inputs.geometry,
                                       lonhit, lathit)
        probability = outputs.randgen.random(npackets)
        v_emit = outputs.surfaceint.v_interp(surftemp, probability)
        v_emit /= outputs.inputs.geometry.planet.radius.value
        
        afactor = outputs.inputs.surfaceinteraction.accomfactor
        v_new = np.sqrt(v_emit**2 * afactor + v_old2 * (1-afactor))
        
    Xtemp[:,4:7] = direction * v_new[:,np.newaxis]

    # Adjust the fractional values
    if outputs.inputs.surfaceinteraction.sticktype == 'temperature dependent':
        lonhit = (np.arctan2(Xtemp[:,1], -Xtemp[:,2])+2*np.pi) % (2*np.pi)
        lathit = np.arcsin(Xtemp[:,3])
        stickcoef = outputs.surfaceint.stickcoef(lonhit, lathit)
        assert np.all(stickcoef <= 1) and np.all(stickcoef >= 0), (
            'Problem with the sticking coefficient')
        Xtemp[:,7] *= (1 - stickcoef)
    elif outputs.inputs.surfaceinteraction.sticktype == 'surface map':
        assert 0
    elif outputs.inputs.surfaceinteraction.stickcoef > 0:
        Xtemp[:,7] *= (1 - outputs.inputs.surfaceinteraction.stickcoef)
    elif outputs.inputs.surfaceinteraction.stickcoef == 0:
        pass
    else:
        assert 0, 'Should not be able to get here.'
        
    # Put new values back in
    Ximpcheck[hitplanet,:] = Xtemp



import numpy as np
import astropy.units as u
from nexoclom.math import random_deviates_1d, random_deviates_2d
from nexoclom.math.distributions import MaxwellianDist, sputdist
from nexoclom.math import Histogram, smooth
import pytest



@pytest.mark.math
def test_random_deviates():
    # Test against gaussian
    nums = np.random.random(3)
    mean = nums[1] * 6 - 3
    sigma = nums[2] * 3
    x0 = np.linspace(-10, 10, 1000)
    f0 = np.exp(-(x0 - mean)**2/sigma**2)

    dev = random_deviates_1d(x0, f0, int(1e7))
    print(dev.mean(), mean)
    print(dev.std(), sigma/np.sqrt(2))
    # assert dev.mean() == pytest.approx(mean, abs=1e-2)
    # assert dev.std() == pytest.approx(sigma/np.sqrt(2), abs=1e-2)

    hist = Histogram(dev, bins=x0)
    hist.histogram /= hist.histogram.mean()
    f0_ = np.exp(-(hist.x - mean)**2/sigma**2)
    f0_ /= f0_.mean()
    print(np.mean((hist.histogram-f0_)**2),np.std((hist.histogram-f0_)**2))
    # assert np.mean((hist.histogram-f0_)**2) == pytest.approx(0, abs=1e-3)
    # assert np.std((hist.histogram-f0_)**2) == pytest.approx(0, abs=1e-3)

    # Test against maxwellian
    v1 = np.linspace(0, 10, 1000)*u.km/u.s
    f1 = MaxwellianDist(v1, 1500.*u.K, 'Na')
    dev = random_deviates_1d(v1, f1, int(1e8))
    hist = Histogram(dev.value, bins=v1.value)
    hist.histogram /= hist.histogram.mean()
    f1_ = MaxwellianDist(hist.x*u.km/u.s, 1500.*u.K, 'Na')
    f1_ /= f1_.mean()
    print(np.mean((hist.histogram-f1_)**2),np.std((hist.histogram-f1_)**2))
    assert np.mean((hist.histogram-f1_)**2) == pytest.approx(0, abs=1e-3)
    assert np.std((hist.histogram-f1_)**2) == pytest.approx(0, abs=1e-3)

if __name__ == '__main__':
    test_random_deviates()


import os
import numpy as np
from scipy import interpolate
import pandas as pd
import pytest
import astropy.units as u
from astropy.time import Time
from nexoclom.solarsystem import SSObject
from nexoclom import Input, __file__ as basefile
from nexoclom.modelcode.source_distribution import surface_distribution
from nexoclom.modelcode.input_classes import SpatialDist
import nexoclom.math as mathMB

import matplotlib.pyplot as plt


# import warnings
# warnings.filterwarnings("error")

basepath = os.path.dirname(basefile)
if __name__ == '__main__':
    inputpath = os.path.join('test_data', 'inputfiles')
else:
    inputpath = os.path.join(basepath, 'modelcode', 'tests', 'test_data', 'inputfiles')


NPACKETS = 100000

def Q_ks(z):
    # Numerical Recipes (3rd edition), Chapter 6.14.12
    j = np.arange(100000) + 1
    # return 2*np.sum((-1.)**(j-1) * np.exp(-2*j**2 * z**2))
    return 1 - np.sqrt(2*np.pi)/z * np.sum(np.exp(-(2*j-1)**2*np.pi**2/(8*z**2)))

class Output_:
    def __init__(self, inputs):
        self.inputs = inputs
        self.npackets = NPACKETS
        self.randgen = np.random.default_rng()
        self.X0 = pd.DataFrame()


spatial_inputs = [{'type': 'surface spot',
                   'longitude': '4.712388',
                   'latitude': '0',
                   'sigma': '0.87266'},
                  {'type': 'uniform'},
                  {'type': 'uniform',
                   'exobase': 1.0,
                   'longitude': '1, 2',
                   'latitude': '0, 1'},
                  {'type': 'uniform',
                   'exobase': 1.0,
                   'longitude': '0, 0',
                   'latitude': '0, 0'}]


def spotmap(lon0, lat0, sigma0):
    spot0 = ((np.sin(lon0)*np.cos(lat0)).value,
             (-np.cos(lon0)*np.cos(lat0)).value,
             (np.sin(lat0)).value)
    longitude = np.linspace(0, 2*np.pi, 361)*u.rad
    latitude = np.linspace(-np.pi/2, np.pi/2, 181)*u.rad
    
    ptsx = np.outer(np.sin(longitude.value), np.cos(latitude.value))
    ptsy = -np.outer(np.cos(longitude.value), np.cos(latitude.value))
    ptsz = -np.outer(np.ones_like(longitude.value), np.sin(latitude.value))
    
    cosphi = ptsx*spot0[0]+ptsy*spot0[1]+ptsz*spot0[2]
    cosphi[cosphi > 1] = 1
    cosphi[cosphi < -1] = -1
    phi = np.arccos(cosphi)
    sourcemap_ = np.exp(-phi/sigma0.value)
    sourcemap = {'longitude': longitude,
                 'latitude': latitude,
                 'map': sourcemap_}
    
    return sourcemap


def ks_test(data, expected):
    p_x = np.sqrt(NPACKETS) + 0.12 + 0.11/np.sqrt(NPACKETS)
    cum_data = mathMB.CumDist(data)
    cum_expect = mathMB.CumDist(data, expected)
    
    D = np.max(np.abs(cum_data.sum - cum_expect.sum))
    p = Q_ks(p_x * D)
    
    if not np.isclose(p, 1):
        plt.plot(cum_data.x, cum_data.sum)
        plt.plot(cum_expect.x, cum_expect.sum)
        plt.show()
        from IPython import embed; embed()
    
    return D, p


@pytest.mark.modelcode
@pytest.mark.parametrize('sparams', spatial_inputs)
def test_surface_distribution(sparams):
    # Full surface uniform distribution
    inputfile = os.path.join(inputpath, 'Spatial.01.input')
    inputs = Input(inputfile)
    inputs.spatialdist = SpatialDist(sparams)
    outputs = Output_(inputs)
    
    surface_distribution(outputs)
    X0 = outputs.X0
    assert X0.shape == (NPACKETS, 6)
    
    longitude, latitude = X0.longitude.values, X0.latitude.values

    if sparams['type'] == 'uniform':
        # longitude
        lon = tuple(map(lambda x:x.value, inputs.spatialdist.longitude))
        expect_lon = np.ones_like(longitude)
        
        if lon[0] < lon[1]:
            expect_lon[(longitude < lon[0]) | (longitude > lon[1])] = 0
        else:
            expect_lon[(longitude > lon[1]) & (longitude < lon[0])] = 0

        # Latitude
        lat = tuple(map(lambda x:np.sin(x.value), inputs.spatialdist.latitude))
        expect_lat = np.ones_like(latitude)
        expect_lat[(np.sin(latitude) < lat[0]) | (np.sin(latitude) > lat[1])] = 0
    elif sparams['type'] == 'surface spot':
        sourcemap = spotmap(inputs.spatialdist.longitude,
                            inputs.spatialdist.latitude,
                            inputs.spatialdist.sigma)
        lon = sourcemap['longitude'].value
        lat = sourcemap['latitude'].value
        
        weight = sourcemap['map'] * np.sin(sourcemap['latitude'])
        expect_lon = np.interp(longitude, lon, np.sum(weight, axis=1))
        expect_lat = np.interp(np.sin(latitude), np.sin(lat),
                               np.sum(sourcemap['map'], axis=0))
        
    else:
        assert False
        
    from IPython import embed; embed()
    import sys; sys.exit()
    

    _, p_lon = ks_test(longitude, expect_lon)
    assert p_lon == pytest.approx(1)

    _, p_lat = ks_test(np.sin(latitude), expect_lat)
    assert p_lat == pytest.approx(1)


if __name__ == '__main__':
    for sparams in spatial_inputs:
        print(sparams)
        test_surface_distribution(sparams)



"""``atomicmass`` - Return the atomic mass of an atom or molecule.

This is really just a wrapper for
`periodictable
<https://periodictable.readthedocs.io/en/latest/index.html>`_
but returns the mass as an `astropy quantity
<http://docs.astropy.org/en/stable/units/index.html>`_.
"""
import periodictable as pt
import astropy.units as u


def atomicmass(species):
    r"""Return the atomic mass of an atom or molecule.

    **Parameters**

    species
        Chemical formula requested species. See `periodictable
        <https://periodictable.readthedocs.io/en/latest/index.html>`_
        for formatting options.

    **Returns**

    The atomicmass of *species* as an astropy quantity with units = AMU
    :math:`(1\, \mathrm{AMU} = 1.660539 \times 10^{-27}\, \mathrm{kg})`.
    If ``periodictable`` returns a ValueError, *None* is returned.
    
    **Examples**
    ::
    
        >>> from nexoclom.atomicdata import atomicmass
        >>> print(atomicmass('Na'))
        22.98977 u
        >>> print(atomicmass('H2O'))
        18.01528 u
        >>> print(atomicmass('X'))
        WARNING: mathMB.atomicmass: X not found
        None
    
    """
    el = [e.symbol for e in pt.elements]
    if species in el:
        atom = eval('pt.' + species)
        mass = atom.mass * u.u
    else:
        try:
            mass = pt.formula(species).mass * u.u
        except ValueError:
            print(f'WARNING: mathMB.atomicmass: {species} not found')
            mass = None

    return mass



import pytest
from nexoclom.utilities.configure import (configure, configure_nexoclom,
                                          configure_solarsystem,
                                          configure_atomicdata)

# @pytest.mark.utilities
# def test_configure():
